from hashlib import pbkdf2_hmac
from base64 import b64encode
from uuid import UUID
from typing import TypeVar, List, Callable, Union, Dict, Type, Optional
from enum import Enum
from abc import abstractmethod, ABC
from io import BytesIO, SEEK_SET, SEEK_END
import random
from datetime import datetime
from dateutil.parser import parse

from pbkdf2 import PBKDF2

from datalogue.errors import DtlError, _enum_parse_error


hardcoded_salt = b"\xc7H\x94\x80l\x9f\x8f\t"
T = TypeVar("T")
R = TypeVar("R")
Json = Union[Dict, str, int, List]


def generate_salt(password: str) -> Union[DtlError, str]:
    """
    Generate a hexadecimal string usable for salting by an hash algorithm based on a simple password

    :param password: a simple password that can be used to generate a hexadecimal string appropriate for salting a hash
    :return: a hexadecimal string, if successful, or DtlError if failed
    """
    if not isinstance(password, str):
        return DtlError("Passwords need to be a string type!")
    salt = pbkdf2_hmac("sha512", password.encode("utf-8"), hardcoded_salt, 100000)

    return salt.hex()


def exists(func, iterable) -> bool:
    for i in iterable:
        if func(i) is True:
            return True
    return False


def map_option(optional_value: Optional[T], function: Callable[[T], R]) -> Optional[R]:
    if optional_value is None:
        return None
    else:
        return function(optional_value)


def _parse_datetime(value: str) -> Union[DtlError, datetime]:
    try:
        datetime_value = parse(value)
        return datetime_value
    except ValueError:
        return DtlError(f"Given value: {value} is not a proper datetime!")


def _parse_uuid(value: Union[str, UUID], error_message: Optional[str] = None) -> Union[DtlError, UUID]:
    if isinstance(value, UUID):
        return value
    try:
        uuid_value = UUID(value)
        return uuid_value
    except ValueError:
        if error_message is None:
            return DtlError(f"Given value: {value} is not a proper UUID!")
        else:
            return DtlError(error_message)


def _parse_list(
    parse_function: Callable[[Json], Union[DtlError, T]]
) -> Callable[[List[Json]], Union[DtlError, List[T]]]:
    """
    Returns another function that can be used to parse a list of json with the specified parse function

    :param parse_function: function to be used for parsing
    :return:
    """

    def parse_concrete(objects: List[Json]) -> Union[DtlError, List[T]]:
        """
        Applies the parse function specified in the parent function to each element of the list of json

        :param objects: list to apply the parse function to
        :return:
        """
        parsed_list = []
        for obj in objects:
            parsed_obj = parse_function(obj)

            if isinstance(parsed_obj, DtlError):
                return parsed_obj
            else:
                parsed_list.append(parsed_obj)

        return parsed_list

    return parse_concrete


def _parse_string_list(objects: List[Json]) -> Union[DtlError, List[str]]:
    parsed_list = []

    for obj in objects:
        if isinstance(obj, str):
            parsed_list.append(obj)
        else:
            return DtlError("The following object is not a string: %s" % obj)

    return parsed_list


def not_implemented() -> NotImplemented:
    return NotImplemented


def random_alpha_numeric_string() -> str:
    alphabet = "".join(random.choice("abcdefghijklmnopqrstuvqxyz") for i in range(5))
    numeric = "".join(random.choice("0123456789") for i in range(3))
    return alphabet + numeric


def random_alpha_numeric_string_with_special_chars() -> str:
    alpha_numeric = random_alpha_numeric_string()
    special_chars = "".join(random.choice("""!#$%&'*+-/=?^_`{|}~""") for i in range(3))
    return alpha_numeric + special_chars


def is_valid_uuid(val):
    uuid_or_not = _parse_uuid(val)
    if isinstance(uuid_or_not, UUID):
        return True
    else:
        return False


class SerializableStringEnum(Enum):
    @staticmethod
    def from_str(enum: Type["SerializableStringEnum"],) -> Callable[[str], Union[DtlError, "SerializableStringEnum"]]:
        """
        Builds a function to parse the string enum `enum`
        :param enum: enum to create a parser for
        :return:
        """

        def inner_sanctum(s: str) -> Union[DtlError, "SerializableStringEnum"]:
            """
            Parses a string and returns the instance of the Enum it corresponds to or a string with an error message

            :param s: string to be parsed
            :return:
            """
            for blob_type in enum:
                if blob_type.value == s:
                    return blob_type

            return enum.parse_error(s)

        return inner_sanctum

    def __repr__(self):
        return f"{self._value_}"

    @staticmethod
    @abstractmethod
    def parse_error(s: str) -> DtlError:
        """
        Returns the error to be used if the parsing fails
        :return: string error
        """
        return s


class ResponseStream(object):
    def __init__(self, request_iterator):
        """
        Class used internally to be able to transform a `requests` iterator into a File like object.

        :param request_iterator: Iterator to pull the data from
        """

        # In memory bytes buffer
        self._bytes = BytesIO()
        # Iterator to use to retrieve data
        self._iterator = request_iterator

        self.closed = self._bytes.closed

    def close(self):
        return self._bytes.close()

    def _load_all(self) -> int:
        """
        Loads all the data into the buffer

        :return: new absolute position in the buffer
        """
        self._bytes.seek(0, SEEK_END)
        for chunk in self._iterator:
            self._bytes.write(chunk)

    def _load_until(self, goal_position: int) -> None:
        """
        Loads the data into the inner buffer until the goal position

        :param goal_position: Byte index to go to
        :return:
        """

        # End of the byte buffer
        current_position = self._bytes.seek(0, SEEK_END)

        # Fill the buffer with more data until we reach the new byte index goal
        while current_position < goal_position:
            try:
                current_position += self._bytes.write(next(self._iterator))
            except StopIteration:
                break

    def tell(self) -> int:
        """
        Current file position, an integer.

        :return:
        """
        return self._bytes.tell()

    def read(self, size=None) -> bytes:
        """
        Read at most size bytes, returned as a bytes object.

        :param size: in bytes to retrieve, if negative reads all
        :return: byte object
        """
        left_off_at = self._bytes.tell()

        # Fill the buffer with new data from the iterator
        if size is None:
            self._load_all()
        else:
            goal_position = left_off_at + size
            self._load_until(goal_position)

        # make sure we are reading from the right position
        self._bytes.seek(left_off_at)

        # return the size bytes from the buffer
        return self._bytes.read(size)

    def seek(self, position, whence=SEEK_SET) -> int:
        """

        :param position:
        :param whence:
        :return:
        """
        if whence == SEEK_END:
            self._load_all()
            return self._bytes.seek(position, whence)
        else:
            return self._bytes.seek(position, whence)


class EncryptionAlgorithm(SerializableStringEnum):
    AES_256 = "AES256"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("encryption algorithm", s))

    @staticmethod
    def encryption_algorithm_from_str(string: str,) -> Union[DtlError, "EncryptionAlgorithm"]:
        return SerializableStringEnum.from_str(EncryptionAlgorithm)(string)


def generate_key(
    password: str, encryption_algorithm: EncryptionAlgorithm = EncryptionAlgorithm.AES_256,
) -> Union[DtlError, str]:
    """
        Generate a key usable by an encryption algorithm based on a simple password

        :param password: a simple password that can be used to generate a key appropriate for encryption
        :param encryption_algorithm: the algorithm that will be used along with the encryption key
        :return: an encryption key as a string, if successful, or DtlError if failed
    """
    if not isinstance(password, str):
        return DtlError("Passwords need to be a string type!")
    if encryption_algorithm == EncryptionAlgorithm.AES_256:
        # Because we just want a mapping of password -> key, we use a hardcoded salt. Use of PBKDF2 is retained because of its useful password -> key generation.
        key = PBKDF2(password, hardcoded_salt).read(32)
        # We'll have to decode the bytestring in UTF-8,
        return b64encode(key).decode("utf-8")
    else:
        return DtlError("Encryption algorithm unknown")



import itertools
from typing import Dict, List, Optional, Union
from uuid import UUID

from datalogue.dtl_utils import _parse_list
from datalogue.errors import DtlError
from datalogue.models.permission import OntologyPermission
from datalogue.models.scope_level import Scope


class TrainingDataColumn:
    def __init__(self, datastore_name: str, path: List[str] = [], count: int = 0):

        if count is not 0 and not isinstance(count, int):
            raise DtlError("count should be an integer in TrainingDataColumn")

        self.datastore_name = datastore_name
        self.path = path
        self.count = count

    def __eq__(self, other: "TrainingDataColumn"):
        if isinstance(self, other.__class__):
            return self.datastore_name == other.datastore_name and self.path == other.path and self.count == other.count
        return False

    @staticmethod
    def _as_payload(training_data_column: "TrainingDataColumn",) -> Union[DtlError, dict]:
        payload = {
            "datastore_name": str(training_data_column.datastore_name),
            "path": list(map(lambda n: str(n), training_data_column.path)),
            "count": training_data_column.count,
        }

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "TrainingDataColumn"]:
        datastore_name = payload.get("datastore_name")
        path = payload.get("path")
        count = payload.get("count")

        if datastore_name is None:
            return DtlError("'datastore_name' not defined in TrainingDataColumn")

        if path is None:
            path = []

        return TrainingDataColumn(datastore_name, path, count)


class OntologyNode:
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        children: List["OntologyNode"] = [],
        id: Optional[UUID] = None,
        total_training_count: int = 0,
        training_data_columns: List[TrainingDataColumn] = [],
    ):

        if not isinstance(name, str):
            raise DtlError("name should be string in OntologyNode")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in OntologyNode")

        if not isinstance(children, List):
            raise DtlError("children should be list in OntologyNode")

        if id is not None and not isinstance(id, UUID):
            raise DtlError("id should be uuid in OntologyNode")

        if total_training_count is not 0 and not isinstance(total_training_count, int):
            raise DtlError("total_training_count should be an integer in OntologyNode")

        self.name = name
        self.description = description
        self.children = children
        self.id = id
        self.total_training_count = total_training_count
        self.training_data_columns = training_data_columns

    def __eq__(self, other: "OntologyNode"):
        if isinstance(self, other.__class__):
            return (
                self.name == other.name
                and self.description == other.description
                and self.children == other.children
                and self.id == other.id
                and self.total_training_count == other.total_training_count
                and self.training_data_columns == other.training_data_columns
            )
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}(id: {self.id}, name: {self.name!r})"

    @staticmethod
    def as_payload(ontology_node: "OntologyNode") -> Union[DtlError, dict]:
        payload = {
            "name": ontology_node.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology_node.children)),
            "training_data_columns": list(
                map(lambda n: TrainingDataColumn._as_payload(n), ontology_node.training_data_columns,)
            ),
            "total_training_count": ontology_node.total_training_count,
        }

        if ontology_node.id is not None:
            payload["node_id"] = str(ontology_node.id)

        if ontology_node.description is not None:
            payload["description"] = ontology_node.description

        return payload

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, "OntologyNode"]:
        id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        children = payload.get("children")
        total_training_count = payload.get("total_training_count")
        training_data_columns = payload.get("training_data_columns")

        if children is None:
            children = []
        else:
            children = _parse_list(OntologyNode.from_payload)(children)

        if training_data_columns is None:
            training_data_columns = []
        else:
            training_data_columns = _parse_list(OntologyNode.from_payload)(training_data_columns)

        if total_training_count is None:
            total_training_count = 0
        return OntologyNode(name, description, children, id, total_training_count, training_data_columns)


class Ontology:
    def __init__(
        self, name: str, description: Optional[str] = None, tree: List[OntologyNode] = [], id: Optional[UUID] = None,
    ):

        if not isinstance(name, str):
            raise DtlError("name should be string in Ontology")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in Ontology")

        if not isinstance(tree, List):
            raise DtlError("tree should be list in Ontology")

        if id is not None and not isinstance(id, UUID):
            raise DtlError("id should be uuid in Ontology")

        self.name = name
        self.description = description
        self.tree = tree
        self.id = id

    def __eq__(self, other: "Ontology"):
        if isinstance(self, other.__class__):
            return (
                self.name == other.name
                and self.description == other.description
                and self.tree == other.tree
                and self.id == other.id
            )
        return False

    def __repr__(self):
        def print_nodes(tree, output=[], level=0):
            for n in tree:
                padding = "   " * level
                if level == 0:
                    output.append(padding + n.name)
                else:
                    output.append(padding + "|___" + n.name)
                for c in n.children:
                    print_nodes([c], output, level=level + 1)
            return output

        first_line = f"Ontology(id: {self.id}, name: {self.name!r}, description: {self.description!r})" + "\n"
        return "\n".join(print_nodes(self.tree, [first_line]))

    def leaves(self) -> List[OntologyNode]:
        def iterate(node: OntologyNode) -> List[OntologyNode]:
            if not node.children:
                return [node]
            else:
                return list(itertools.chain(*map(lambda n: iterate(n), node.children)))

        return list(itertools.chain(*map(lambda n: iterate(n), self.tree)))

    @staticmethod
    def _as_payload(ontology: "Ontology") -> Union[DtlError, dict]:
        payload = {
            "name": ontology.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology.tree)),
        }

        if ontology.description is not None:
            payload["description"] = ontology.description

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Ontology"]:
        id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        tree = payload.get("tree")
        if tree is None:
            tree = []
        else:
            tree = _parse_list(OntologyNode.from_payload)(payload["tree"])
        return Ontology(name, description, tree, id=id)

    @staticmethod
    def _create_body_for_sharing(target_id: UUID, target_type: Scope, permission: OntologyPermission) -> Dict:

        if not isinstance(target_type, Scope):
            return DtlError(
                f"'{target_type}' should be a type of '{Scope.__name__}' and must be in: {list(map(str, Scope))}"
            )

        if not isinstance(permission, OntologyPermission):
            return DtlError(
                f"'{permission}' should be a type of '{OntologyPermission.__name__}' and must be in: {list(map(str, OntologyPermission))}"
            )

        body = {}

        body["organizations"] = None
        body["groups"] = None
        body["users"] = None

        if target_type == Scope.User:
            body["users"] = {str(target_id): permission.value}
        elif target_type == Scope.Group:
            body["groups"] = {str(target_id): permission.value}
        elif target_type == Scope.Organization:
            body["organizations"] = {str(target_id): permission.value}
        return body

    @staticmethod
    def _create_body_for_unsharing(target_id: UUID, target_type: Scope, permission: str) -> Dict:

        if not isinstance(target_type, Scope):
            return DtlError(
                f"'{target_type}' should be a type of '{Scope.__name__}' and must be in: {list(map(str, Scope))}"
            )

        body = {}

        body["organizations"] = None
        body["groups"] = None
        body["users"] = None

        if target_type == Scope.User:
            body["users"] = {str(target_id): permission}
        elif target_type == Scope.Group:
            body["groups"] = {str(target_id): permission}
        elif target_type == Scope.Organization:
            body["organizations"] = {str(target_id): permission}
        return body


class DataRef:
    def __init__(self, node_id: Union[str, UUID], path_list: List[List[str]]):
        self.node_id = node_id
        self.path_list = path_list


class Mapping:
    def __init__(self, name: str, path: str):
        self.name = name
        self.path = path

    def __repr__(self):
        return f"{self.__class__.__name__}(name: {self.name}, path: {self.path})"

    def __eq__(self, other):
        if isinstance(other, Mapping):
            return self.name == other.name and self.path == other.path
        return False

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, "Mapping"]:
        name = payload.get("name")
        if not isinstance(name, str):
            return DtlError("'name' parameter is required for Mapping")

        path = payload.get("path")
        if not isinstance(path, str):
            return DtlError("'path' parameter is required for Mapping")

        return Mapping(name, path)


class ClassMapping:
    def __init__(
        self, class_id: UUID, source_mappings: List[Mapping], destination_mappings: List[Mapping],
    ):
        self.class_id = class_id
        self.source_mappings = source_mappings
        self.destination_mappings = destination_mappings

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(class_id: {self.class_id}, source_mappings: {self.source_mappings!r}, "
            f"destination_mappings: {self.destination_mappings!r})"
        )

    def __eq__(self, other):
        if isinstance(other, ClassMapping):
            return (
                self.class_id == other.class_id
                and self.source_mappings == other.source_mappings
                and self.destination_mappings == other.destination_mappings
            )
        return False

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, "ClassMapping"]:
        class_id = payload.get("classId")
        if not isinstance(class_id, str):
            return DtlError("'classId' should be a UUID in ClassMapping")
        class_id = UUID(class_id)

        source_mappings = payload.get("sourceMappings")
        if not isinstance(source_mappings, List):
            return DtlError("'sourceMappings' should be a List in ClassMapping")
        source_mappings = _parse_list(Mapping.from_payload)(source_mappings)

        destination_mappings = payload.get("destinationMappings")
        if not isinstance(destination_mappings, List):
            return DtlError("'destinationMappings' should be a List in ClassMapping")
        destination_mappings = _parse_list(Mapping.from_payload)(destination_mappings)

        return ClassMapping(class_id, source_mappings, destination_mappings)



from typing import List, Optional, Union
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.job import Job, _job_from_payload
from datalogue.dtl_utils import _parse_list
from datalogue.errors import DtlError
from uuid import UUID


class _JobsClient:
    """
    Client to interact with the Scheduled pipelines
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def list(
        self, page: int = 1, item_per_page: int = 25, pipeline_id: Optional[UUID] = None
    ) -> Union[DtlError, List[Job]]:
        """
        List jobs

        :param page: page to be retrieved
        :param item_per_page: number of jobs to be put in a page
        :param pipeline_id: optional pipeline id used to retrieve related jobs.
        :return: Returns a List of all the available Jobs or an error message as a string
        """

        if pipeline_id:
            endpoint = f"/jobs?page={page}&size={item_per_page}&stream-id={str(pipeline_id)}"
        else:
            endpoint = f"/jobs?page={page}&size={item_per_page}"

        res = self.http_client.make_authed_request(self.service_uri + endpoint, HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_job_from_payload)(res)

    def cancel(self, job_id: UUID) -> Union[DtlError, Job]:
        """"
        Cancel a Job given job_id.

        :param job_id:
        :return: Returns the Job object canceled or an error
        """

        res = self.http_client.make_authed_request(self.service_uri + f"/jobs/{str(job_id)}/cancel", HttpMethod.POST)

        if isinstance(res, DtlError):
            return res

        return _job_from_payload(res)

    def get(self, job_id: UUID) -> Union[DtlError, Job]:
        """"
        Get a Job given job_id.

        :param job_id:
        :return: Returns the Job object or an error
        """

        res = self.http_client.make_authed_request(self.service_uri + f"/jobs/{str(job_id)}/status", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _job_from_payload(res)



from typing import List, Union
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_string_list


class AppendIndexToLabel(Transformation):
    """
    Adds indexes to all the nodes that are present at the same path to be able to distinguish them
    """

    type_str = "AppendIndexToLabel"

    def __init__(self, path: List[str]):
        """
        :param path: Adds indexes to all the nodes at the same path
        """
        Transformation.__init__(self, AppendIndexToLabel.type_str)
        self.path = path

    def __eq__(self, other: "AppendIndexToLabel"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"AppendIndexToLabel(path: {self.path})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "AppendIndexToLabel"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return AppendIndexToLabel(path)



from datalogue.dtl import *
from datalogue.credentials import *
from datalogue.errors import *
from datalogue.models.classifier import *
from datalogue.models.credentials import *
from datalogue.models.datastore import *
from datalogue.models.datastore_collection import *
from datalogue.models.job import *
from datalogue.models.metrics import *
from datalogue.models.ontology import *
from datalogue.models.organization import *
from datalogue.models.permission import *
from datalogue.models.regex import *
from datalogue.models.s3encryption import *
from datalogue.models.schema import *
from datalogue.models.pipeline import *
from datalogue.models.tag import *
from datalogue.models.model import *
from datalogue.models.transformations import *
from datalogue.models.transformations.structure import *
from datalogue.models.transformations.math import *
from datalogue.models.transformations.obfuscate import *
from datalogue.models.transformations.commons import *
from datalogue.models.version import *
from datalogue.dtl_utils import *
from datalogue.models.pipeline_template import *



from typing import List, Union, Optional

from datalogue.models.transformations import Transformation
from datalogue.anduin.models.datastore import Datastore
from datalogue.errors import DtlError

from datetime import datetime
from datalogue.errors import _enum_parse_error
from datalogue.dtl_utils import SerializableStringEnum
from dateutil.parser import parse

from uuid import UUID


class StreamCollection:
    def __init__(
        self, transformations: List[Transformation], pipelines: List["StreamCollection"], target: Datastore,
    ):
        self.transformations = transformations
        self.pipelines = pipelines
        self.target = target

    def _as_payload(self) -> Union[DtlError, dict]:
        return {
            "transformations": list(map(lambda s: s._as_payload(), self.transformations)),
            "pipelines": list(map(lambda s: s._as_payload(), self.pipelines)),
            "target": self.target._as_payload(),
        }


class Stream:
    def __init__(self, datastore: Datastore, stream_collection: List[StreamCollection]):
        self.datastore = datastore
        self.stream_collection = stream_collection

    def _as_payload(self):
        return {
            "source": self.datastore._as_payload(),
            "pipelines": list(map(lambda s: s._as_payload(), self.stream_collection)),
        }

    @staticmethod
    def simple_stream(datastore: Datastore, transformations: List[Transformation], target: Datastore):
        return Stream(datastore, [StreamCollection(transformations, [], target)])


class StreamState(SerializableStringEnum):
    Succeeded = "Succeeded"
    Failed = "Failed"
    Running = "Running"
    Defined = "Defined"
    Unknown = "Unknown"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("stream state", s))

    @staticmethod
    def stream_state_from_str(string: str) -> Union[DtlError, "StreamState"]:
        return SerializableStringEnum.from_str(StreamState)(string)


class StreamStatus:
    def __init__(
        self, stream_id: UUID, state: StreamState, timestamp: datetime, details: Optional[str] = None,
    ):
        self.stream_id = stream_id
        self.state = state
        self.timestamp = timestamp
        self.details = details

    def __repr__(self):
        return f"StreamStatus(id: {self.stream_id}, state: {self.state.value}, ts: {self.timestamp}, details: {self.details})"

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, "StreamStatus"]:
        stream_id = payload.get("id")
        if stream_id is None:
            return DtlError("'stream_id' needs to be defined!")

        ts = payload.get("timestamp")
        if ts is None:
            return DtlError("'timestamp' needs to be defined!")
        else:
            try:
                ts = parse(ts)
            except ValueError:
                return DtlError("'timestamp' could not be parsed as a valid date")

        stream_state = payload.get("status")
        if stream_state is None:
            return DtlError("'status' needs to be defined!")
        else:
            stream_state = StreamState.stream_state_from_str(stream_state)
            if isinstance(stream_state, DtlError):
                return stream_state

        return StreamStatus(UUID(stream_id), stream_state, ts, payload["details"])



from typing import List, Union, Optional
from uuid import UUID
from datalogue.errors import DtlError, _invalid_parameter_error, _property_not_found
from datalogue.models.transformations.commons import (
    Transformation,
    _array_from_dict,
    RegexTransformation,
)


class MoveByRegex(RegexTransformation):
    """
    Finds all nodes that has a label matching to the given regex 
    and moves them as children of a new parent identified by the given to path
    """

    type_str = "MoveByRegex"

    def __init__(self, regex: Optional[str], regex_id: Optional[UUID], to: List[str]):
        """
        :param regex: a string input to find which nodes should be split into two
        :param to: array of string
        """

        RegexTransformation.__init__(self, regex, regex_id, MoveByRegex.type_str)
        self.to = to
        self.regex = regex
        self.regex_id = regex_id

    def __eq__(self, other: "MoveByRegex"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"MoveByRegex(regex: {self.regex}, regex_id: {self.regex_id}, to: {'.'.join(self.to)})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        if self.regex is not None:
            base["regex"] = self.regex
        if self.regex_id is not None:
            base["regexId"] = str(self.regex_id)
        base["to"] = self.to
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "MoveByRegex"]:
        regex = json.get("regex")
        regex_id = json.get("regexId")

        if regex is None and regex_id is None:
            return _property_not_found("neither 'regex' nor 'regexId", json)

        to = _array_from_dict(json, MoveByRegex.type_str, "to")
        if isinstance(to, DtlError):
            return to

        return MoveByRegex(regex, regex_id, to)



import abc
from typing import List, Optional, Union
from datalogue.errors import DtlError, _property_not_found, _invalid_property_type, _enum_parse_error
from datetime import datetime
from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.dtl_utils import _parse_list, SerializableStringEnum


class NodeDescription:
    """
    Description of a node to be added to the tree.
    """

    def __init__(
        self, path: List[str], value: Union[str, bool, int, float, datetime], value_type: DataType,
    ):
        """
        :param path: path of the node to be created
        :param value: value to be set for the node
        :param value_type: type of the value to be used
        """
        self.path = path
        self.value = value
        self.value_type = value_type

    def __eq__(self, other: "NodeDescription"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"NodeDescription(path: {'.'.join(self.path)}, value: {self.value!r}, type: {self.value_type!r})"

    def _as_payload(self) -> Union[DtlError, dict]:
        value = self.value
        if isinstance(value, datetime):
            value = value.isoformat()
        else:
            value = str(value)

        return {"path": self.path, "value": value, "valueType": self.value_type.value}

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "NodeDescription"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)
        elif not isinstance(path, list):
            return _invalid_property_type("path", "List[str]", json)

        value = json.get("value")
        if value is None:
            return _property_not_found("value", json)
        elif not isinstance(value, str):
            return _invalid_property_type("value", "str", json)

        value_type = json.get("valueType")
        if value_type is None:
            return _property_not_found("valueType", json)
        else:
            value_type = DataType.from_str(value_type)
            if isinstance(value_type, DtlError):
                return value_type

        return NodeDescription(path, value, value_type)


class Add(Transformation):
    """
    Adds a node to the graph

    Some edge cases:
        - If the specified path's parent do not exists, the node will not be created
        - If the cast cannot happen for the given value, the node will node be created
    """

    type_str = "Add"

    def __init__(self, nodes: List[NodeDescription]):
        """
        :param nodes: List of nodes to be added to the tree
        """
        Transformation.__init__(self, Add.type_str)
        self.nodes = nodes

    def __eq__(self, other: "Add"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Add(nodes: {self.nodes})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["nodes"] = list(map(lambda x: x._as_payload(), self.nodes))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Add"]:
        nodes = json.get("nodes")
        if nodes is None:
            return _property_not_found("nodes", json)

        nodes = _parse_list(NodeDescription._from_payload)(nodes)
        if isinstance(nodes, DtlError):
            return nodes

        return Add(nodes)


class TimeFormat(abc.ABC):
    type_field = "type"

    def __init__(self, time_format_type: str):
        self.type = time_format_type
        super().__init__()

    def __eq__(self, other: "TimeFormat"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _base_payload(self) -> dict:
        base = [(TimeFormat.type_field, self.type)]
        return dict(base)

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the custom type in its dictionary construction

        :return:
        """


class EpochSeconds(TimeFormat):
    type_str = "EpochSeconds"

    def __init__(self):
        TimeFormat.__init__(self, EpochSeconds.type_str)

    def __repr__(self):
        return EpochSeconds.type_str

    def _as_payload(self) -> dict:
        base = self._base_payload()
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "EpochSeconds"]:
        return EpochSeconds()


class EpochMillis(TimeFormat):
    type_str = "EpochMillis"

    def __init__(self):
        TimeFormat.__init__(self, EpochMillis.type_str)

    def __repr__(self):
        return EpochMillis.type_str

    def _as_payload(self) -> dict:
        base = self._base_payload()
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "EpochMillis"]:
        return EpochMillis()

def _time_format_from_payload(json: dict) -> Union[DtlError, "TimeFormat"]:
    type_str = json.get(TimeFormat.type_field)
    if type_str is None:
        return _property_not_found("type", json)
    if type_str == EpochSeconds.type_str:
        time_format = EpochSeconds._from_payload(json)
    elif type_str == EpochMillis.type_str:
        time_format = EpochMillis._from_payload(json)
    else:
        return DtlError("'%s' time format is not handled by the SDK" % type_str)
    return time_format


class CustomType(abc.ABC):
    type_field = "type"

    def __init__(self, custom_type: str):
        self.type = custom_type
        super().__init__()

    def __eq__(self, other: "CustomType"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _base_payload(self) -> dict:
        base = [(CustomType.type_field, self.type)]
        return dict(base)

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the custom type in its dictionary construction

        :return:
        """


class RootName(CustomType):
    type_str = "RootName"

    def __init__(self):
        CustomType.__init__(self, RootName.type_str)

    def __repr__(self):
        return RootName.type_str

    def _as_payload(self) -> dict:
        base = self._base_payload()
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RootName"]:
        return RootName()


class RecordProcessingTimestamp(CustomType):
    type_str = "RecordProcessingTimestamp"

    def __init__(self, time_format: Optional["TimeFormat"] = None):
        CustomType.__init__(self, RecordProcessingTimestamp.type_str)
        self.time_format = time_format

    def __repr__(self):
        return f"RecordProcessingTimestamp(timeFormat :{self.time_format!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        if self.time_format is not None:
            base["timeFormat"] = self.time_format._as_payload()
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RecordProcessingTimestamp"]:
        time_format_dict = json.get("timeFormat")
        if time_format_dict is not None:
            time_format = _time_format_from_payload(time_format_dict)
        else:
            time_format = None
        return RecordProcessingTimestamp(time_format)


class StringConstant(CustomType):
    type_str = "StringConstant"

    def __init__(self, custom_value: str):
        CustomType.__init__(self, StringConstant.type_str)
        self.custom_value = custom_value

    def __repr__(self):
        return f"StringConstant(customValue: {self.custom_value})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["customValue"] = self.custom_value
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "StringConstant"]:
        custom_value = json.get("customValue")
        if custom_value is None:
            return _property_not_found("customValue", json)
        return StringConstant(custom_value)


class JobRunAt(CustomType):
    type_str = "JobRunAt"

    def __init__(self):
        CustomType.__init__(self, JobRunAt.type_str)

    def _as_payload(self) -> dict:
        base = self._base_payload()
        return base

    def __repr__(self):
        return JobRunAt.type_str

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "JobRunAt"]:
        return JobRunAt()


def _custom_type_from_payload(json: dict) -> Union[DtlError, CustomType]:
    type_str = json.get(CustomType.type_field)
    if type_str is None:
        return _property_not_found("type", json)
    if type_str == RootName.type_str:
        custom_type = RootName._from_payload(json)
    elif type_str == RecordProcessingTimestamp.type_str:
        custom_type = RecordProcessingTimestamp._from_payload(json)
    elif type_str == StringConstant.type_str:
        custom_type = StringConstant._from_payload(json)
    elif type_str == JobRunAt.type_str:
        custom_type = JobRunAt._from_payload(json)
    else:
        return DtlError("'%s' custom type is not handled by the SDK" % type_str)
    return custom_type


class NodeWithCustomValue:
    """
    Description of a node encapsulating a custom value to be added to the tree.
    """

    def __init__(self, path: List[str], custom_type: CustomType):
        """
        :param path: path of the node to be created
        :param custom_type: value to be set for the node
        """
        self.path = path
        self.custom_type = custom_type

    def __eq__(self, other: "NodeWithCustomValue"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"NodeWithCustomValue(path: {'.'.join(self.path)}, custom_type: {self.custom_type!r})"

    def _as_payload(self) -> Union[DtlError, dict]:
        return {"path": self.path, "customType": self.custom_type._as_payload()}

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "NodeWithCustomValue"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)
        elif not isinstance(path, list):
            return _invalid_property_type("path", "List[str]", json)
        custom_type_dict = json.get("customType")
        if custom_type_dict is None:
            return _property_not_found("customType", json)
        custom_type = _custom_type_from_payload(custom_type_dict)
        if isinstance(custom_type, DtlError):
            return custom_type
        return NodeWithCustomValue(path, custom_type)


class AddWithCustomValue(Transformation):
    """
    Adds a field to the record with a custom value.
    """

    type_str = "AddWithCustomValue"

    def __init__(self, nodes: List[NodeWithCustomValue]):
        """
        :param nodes: List of nodes to be added to the tree
        """
        Transformation.__init__(self, AddWithCustomValue.type_str)
        self.nodes = nodes

    def __eq__(self, other: "AddWithCustomValue"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"AddWithCustomValue(nodes: {self.nodes})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["nodes"] = list(map(lambda x: x._as_payload(), self.nodes))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "AddWithCustomValue"]:
        nodes = json.get("nodes")
        if nodes is None:
            return _property_not_found("nodes", json)

        nodes = _parse_list(NodeWithCustomValue._from_payload)(nodes)
        if isinstance(nodes, DtlError):
            return nodes

        return AddWithCustomValue(nodes)



from uuid import UUID

from datalogue.clients.http import _HttpClient, List, Union, HttpMethod, Optional
from datalogue.errors import DtlError
from datalogue.models.permission import (
    Permission,
    SharePermission,
    ObjectType,
    UnsharePermission,
)
from datalogue.models.scope_level import Scope
from datalogue.models.tag import Tag
from datalogue.dtl_utils import _parse_list


class _TagClient:
    """
    Client to interact with the Tags
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, tag: Tag) -> Union[DtlError, List[Tag]]:
        """
        Creates a tag
        :param tag: A Tag object to create
        :return: Returns created Tag object if successful, or DtlError if failed
        """
        payload = {"name": tag.name}
        res = self.http_client.make_authed_request(self.service_uri + "/tags", HttpMethod.POST, [payload])

        if isinstance(res, DtlError):
            return res

        return _parse_list(Tag._from_payload)(res)[0]

    def update(self, old_name: str, name: Optional[str]) -> Union[DtlError, Tag]:
        """
          Updates a tag

          :param id: id of the tag to be updated
          :param name: the updated name to be applied
          :return: Returns Tag object if successful, or DtlError if failed
        """
        payload = {}
        if name is not None:
            payload["name"] = name

        searched_tag_rsp = self.search(old_name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/{searched_tag_rsp.id}", HttpMethod.PUT, payload
        )

        if isinstance(res, DtlError):
            return res

        return Tag._from_payload(res)

    def get(self, name: str) -> Union[DtlError, Tag]:
        """
            Retrieve a tag locally by its id.

            :param name: the name of the tag to be retrieved locally
            :return: the tag if successful, or a DtlError if failed
        """
        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        res = self.http_client.make_authed_request(self.service_uri + f"/tags/{searched_tag_rsp.id}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return Tag._from_payload(res)

    def search(self, name: str) -> Union[DtlError, Tag]:
        """
            Search a tag locally by its name.

            :param name: the name of the tag to be searched locally
            :return: the tag if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(self.service_uri + f"/tags/{name}/search", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return Tag._from_payload(res)

    def delete(self, name: str) -> Union[DtlError, bool]:
        """
            Delete a tag.

            :param name: the name of the tag to be deleted
            :return: Returns True if successful, or DtlError if failed
        """
        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return True

        res = self.http_client.make_authed_request(self.service_uri + f"/tags/{searched_tag_rsp.id}", HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res

        return True

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Tag]]:
        """
            List all tags.

            :param page: page to be retrieved
            :param item_per_page: number of items to be put in a page
            :return: Returns a List of all the available tags if successful, or DtlError if failed
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags?page={page}&size={item_per_page}", HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(Tag._from_payload)(res)

    def share(
        self, name: str, target_type: Scope, target_id: UUID, permission: Permission
    ) -> Union[SharePermission, DtlError]:
        """
            Share a tag.

            :param name: name of the tag to be shared
            :param target_type: Scope (`User`, `Group` or `Organization`) with whom you want to share the tag. It can be User, Group or Organization.
            :param target_id: id of the User, Group or Organization you want to share with (depending on the target_type param)
            :param permission: Permission (`Read`, `Write`, or `Share`) to be shared for a tag. It can be Read, Write or Share.
            :return: Returns a SharePermission for the specific tag if successful, or DtlError if failed
        """

        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        url = f"/scout/tags/{searched_tag_rsp.id}/shares"
        params = {
            "targetType": str(target_type.value),
            "targetId": str(target_id),
            "permission": str(permission.value),
        }
        res = self.http_client.execute_authed_request(url, HttpMethod.POST, params=params)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return SharePermission(object_type=ObjectType.Tag).from_payload(res.json())
        else:
            return DtlError(res.text)

    def unshare(
        self, name: str, target_type: Scope, target_id: UUID, permission: Permission
    ) -> Union[UnsharePermission, DtlError]:
        """
            Unshare a tag.

            :param name: name of the tag to be unshared
            :param target_type: Scope (`User`, `Group` or `Organization`) with whom you want to unshare the tag. It can be User, Group or Organization.
            :param target_id: id of the User, Group or Organization you want to unshare with (depending on the target_type param)
            :param permission: Permission (`Read`, `Write`, or `Share`) to be unshared for a tag. It can be Read, Write or Share.
            :return: Returns an UnsharePermission if successful, or DtlError if failed
        """

        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        url = f"/scout/tags/{searched_tag_rsp.id}/shares"
        params = {
            "targetType": str(target_type.value),
            "targetId": str(target_id),
            "permission": str(permission.value),
        }

        res = self.http_client.execute_authed_request(url, HttpMethod.DELETE, params=params)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return UnsharePermission(object_type=ObjectType.Tag).from_payload(res.json())
        else:
            return DtlError(res.text)

    def get_shares(self, name: str) -> Union[DtlError, List[SharePermission]]:
        """
            List sharings of a tag.

            :param name: name of the tag to get its shared list
            :return: Returns a List of user shares for the specific tag if successful, or DtlError if failed
        """
        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/{searched_tag_rsp.id}/shares", HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(SharePermission(object_type=ObjectType.Tag).from_payload)(res)



from typing import Union
from datalogue.auth.authentication import (
    Authentications,
    _authentication_ref_from_payload,
    AuthenticationsReference,
)
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.errors import DtlError


class _TestAuthenticationClient:
    """
    Client to interact with the Authentication Schemes
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def create(self, authentication_def: Authentications) -> Union[DtlError, AuthenticationsReference]:
        """
        For creating an LDAP object for test cases
        """
        res = self.http_client.make_authed_request(
            "/user-provider-configs?only-validate=false", HttpMethod.POST, authentication_def._as_payload(),
        )
        if isinstance(res, DtlError):
            return res
        return _authentication_ref_from_payload(res)



from typing import Optional, Union, List, Dict, Tuple, Iterator

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list, _parse_string_list
from uuid import UUID


class ProcessedSourceWithStatus:
    def __init__(self, source_name: str, job_id: UUID, status: str):
        self.source_name = source_name
        self.job_id = job_id
        self.status = status

    def __repr__(self):
        return f"(source_name: {self.source_name}, " f"job_id: {self.job_id}, " f"status: {self.status})"


def _parse_processed_source_with_status(json: dict,) -> Union[ProcessedSourceWithStatus, ProcessedSourceWithStatus]:
    source_name = json.get("sourceName")
    if source_name is None:
        return DtlError("'sourceName' for a processed source should be defined")

    job_id = json.get("jobId")
    if job_id is None:
        return DtlError("'jobId' for a processed source should be defined")

    status = json.get("status")
    if status is None:
        return DtlError("'status' for a processed source should be defined")

    return ProcessedSourceWithStatus(source_name, job_id, status)


class _ProcessedSourcesClient:
    """
    Client to get processed sources list for different ids
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    """
    Return the files already processed by the specified job
    """

    def get_by_job_id(self, job_id: Union[str, UUID]) -> List[str]:
        url = f"{self.service_uri}/jobs/{str(job_id)}/sources/lock"
        res = self.http_client.execute_authed_request(url, HttpMethod.GET)
        if isinstance(res, DtlError):
            return res
        return _parse_string_list(res.json())

    """
    Return the files already processed by the specified connection/credential
    """

    def get_by_credential_id(self, credential_id: Union[str, UUID]):
        url = f"{self.service_uri}/credentials/{str(credential_id)}/sources/lock"
        res = self.http_client.execute_authed_request(url, HttpMethod.GET)
        if isinstance(res, DtlError):
            return res
        return _parse_list(_parse_processed_source_with_status)(res.json())

    """
    Return the files already processed by the specified pipeline
    """

    def get_by_stream_id(self, stream_id: Union[str, UUID]):
        url = f"{self.service_uri}/streams/{str(stream_id)}/sources/lock"
        res = self.http_client.execute_authed_request(url, HttpMethod.GET)
        if isinstance(res, DtlError):
            return res
        return _parse_list(_parse_processed_source_with_status)(res.json())



from typing import List, Union, Optional
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_list
from abc import ABC, abstractmethod
from uuid import UUID


class RecognizeEntities(Transformation):
    """
    RecognizeEntities: Runs all data nodes through the entity recognition service using NER model and passes the classification results to the next stage. Since entity recognition process goes through the value of the node and analyzes individual tokens, one node may have several classification tags attached to it.
    """

    type_str = "RecognizeEntities"

    def __init__(self, model_id: UUID, fieldsToTarget: Optional[List[List[str]]] = None):
        """
        Builds an entity recognition transformation

        :param include_classes: augments resulting ADGs with xyz_#DTL_class nodes containing ontology class for corresponding xyz nodes.
        :param include_scores:  augments resulting ADGs with xyz_#DTL_score nodes containing classification score for corresponding xyz nodes.
        """
        Transformation.__init__(self, RecognizeEntities.type_str)
        self.model_id = model_id
        self.fieldsToTarget = fieldsToTarget

    def __eq__(self, other: "RecognizeEntities"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"RecognizeEntities(model_id: {self.model_id}, fieldsToTarget = {self.fieldsToTarget})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["modelId"] = str(self.model_id)
        if self.fieldsToTarget is not None:
            base["fieldsToTarget"] = self.fieldsToTarget
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RecognizeEntities"]:
        if json.get(Transformation.type_field) != RecognizeEntities.type_str:
            return DtlError(f"Type of the transformation is not RecognizeEntities! Payload: {str(json)}")

        model_id = UUID(json.get("modelId"))
        fieldsToTarget = json.get("fieldsToTarget")
        return RecognizeEntities(model_id, fieldsToTarget)



from abc import ABC, abstractmethod
from typing import Dict, List, Set, Optional, Union, Tuple, Callable, Iterator
from dateutil.parser import parse
from datalogue.models.ontology import OntologyNode
from datalogue.models.organization import User
from uuid import UUID
import itertools
import pandas as pd
import datetime
from collections import namedtuple
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.dtl_utils import (
    _parse_list,
    _parse_string_list,
    SerializableStringEnum,
    _parse_uuid,
    map_option,
)


class TrainingData(SerializableStringEnum):
    Synthesized = "Synthesized"
    Original = "Original"


class ModelType(SerializableStringEnum):
    Cbc = "cbc"
    Cfc = "cfc"
    Ner = "ner"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("model type", s))

    def __str__(self):
        return str(self.value)


def model_type_from_str(string: str) -> Union[DtlError, ModelType]:
    return SerializableStringEnum.from_str(ModelType)(string)


class TrainingStatusType(SerializableStringEnum):
    success = "success"
    failure = "failure"
    in_progress = "in-progress"
    requested = "requested"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("training status type", s))


def training_status_type_from_str(string: str) -> Union[DtlError, TrainingStatusType]:
    return SerializableStringEnum.from_str(TrainingStatusType)(string)


class OrderList(SerializableStringEnum):
    asc = "asc"
    desc = "desc"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("order list", s))


def order_list_from_str(string: str) -> Union[DtlError, OrderList]:
    return SerializableStringEnum.from_str(OrderList)(string)


class DataRef:
    def __init__(self, node_id: Union[str, UUID], path_list: List[List[str]]):
        self.node_id = node_id
        self.path_list = path_list


class TrainingEndReason:
    def __init__(self, reasonType: str):
        self.reason_type = reasonType

    def __eq__(self, other: "TrainingEndReason"):
        if isinstance(self, other.__class__):
            return self.reason_type == other.reason_type
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["_type"] = self.reason_type
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "TrainingEndReason"]:
        """
            Builds a TRainingEndReason object from a dictionary,

            :param json: dictionary parsed from json
            :return:
            """
        reasonType = json.get("_type")
        if not isinstance(reasonType, str):
            return DtlError("An '_type' field is missing")

        return TrainingEndReason(reasonType)


class ConfusionMatrix:
    def __init__(self, index: List[str], matrix: List[List[int]]):
        self.index = index
        self.matrix = matrix

    def __eq__(self, other: "ConfusionMatrix"):
        if isinstance(self, other.__class__):
            return self.index == other.index and self.matrix == other.matrix
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["index"] = self.index
        base["matrix"] = self.matrix
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ConfusionMatrix"]:
        """
        Builds a confusion Matrix object from a dictionary,

        :param json: dictionary parsed from json
        :return:
        """
        index = json.get("index")
        if not isinstance(index, List):
            return DtlError("An 'index' field is missing")

        matrix = json.get("matrix")
        if not isinstance(matrix, List):
            return DtlError("An 'matrix' field is missing")

        return ConfusionMatrix(index, matrix)


class MetricsDetails:
    def __init__(
        self,
        timestamp: str,
        metrics_type: str,
        epoch: int,
        loss: float,
        accuracy: float,
        precision: Dict[str, float],
        recall: Dict[str, float],
        f1_score: Dict[str, float],
        tpr: Optional[float] = None,
        fpr: Optional[float] = None,
        conf_matrix: Optional[ConfusionMatrix] = None,
    ):
        self.tpr = tpr
        self.fpr = fpr
        self.timestamp = timestamp
        self.metrics_type = metrics_type
        self.epoch = epoch
        self.loss = loss
        self.accuracy = accuracy
        self.precision = precision
        self.recall = recall
        self.f1_score = f1_score
        self.conf_matrix = conf_matrix

    def __eq__(self, other: "MetricsDetails"):
        if isinstance(self, other.__class__):
            return (
                self.tpr == other.tpr
                and self.fpr == other.fpr
                and self.timestamp == other.timestamp
                and self.metrics_type == other.metrics_type
                and self.epoch == other.epoch
                and self.loss == other.loss
                and self.accuracy == other.accuracy
                and self.precision == other.precision
                and self.recall == other.recall
                and self.f1_score == other.f1_score
                and self.conf_matrix == other.conf_matrix
            )
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["tpr"] = self.tpr
        base["fpr"] = self.fpr
        base["timestamp"] = self.timestamp
        base["metricsType"] = self.metrics_type
        base["epoch"] = self.epoch
        base["loss"] = self.loss
        base["accuracy"] = self.accuracy
        base["precision"] = self.precision._as_payload()
        base["recall"] = self.recall._as_payload()
        base["f1Score"] = self.f1_score._as_payload()

        if self.conf_matrix is not None:
            base["confusionMatrix"] = self.conf_matrix._as_payload()
        else:
            base["confusionMatrix"] = None
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "MetricsDetails"]:
        """
        Builds a MetricsDetails object from a dictionary,

        :param json: dictionary parsed from json
        :return:
        """

        tpr = json.get("tpr")
        if not isinstance(tpr, float):
            tpr = None

        fpr = json.get("fpr")
        if not isinstance(fpr, float):
            fpr = None

        timestamp = json.get("timestamp")
        if not isinstance(timestamp, str):
            return DtlError("A 'timestamp' field is missing")

        metrics_type = json.get("type")
        if not isinstance(metrics_type, str):
            return DtlError("A 'type' field is missing")

        epoch = json.get("epoch")
        if not isinstance(epoch, int):
            epoch = None

        loss = json.get("loss")
        if not isinstance(loss, float):
            return DtlError("A 'loss' field is missing")

        accuracy = json.get("accuracy")
        if not isinstance(accuracy, float):
            return DtlError("An 'accuracy' field is missing")

        precision = json.get("precision")
        if precision is None:
            return DtlError("'Precision' field is missing")
        elif not isinstance(precision, dict):
            return DtlError("'Precision' contains wrong data type it should be dictionary")

        recall = json.get("recall")
        if recall is None:
            return DtlError("'Recall' field is missing")
        elif not isinstance(recall, dict):
            return DtlError("'Recall' contains wrong data type it should be dict")

        f1_score = json.get("f1Score")
        if f1_score is None:
            return DtlError("'F1Score' field is missing")
        elif not isinstance(f1_score, dict):
            return DtlError("'F1Score' contains wrong data type it should be dict")

        conf_matrix = json.get("confusionMatrix")
        if conf_matrix is not None:
            conf_matrix = ConfusionMatrix._from_payload(conf_matrix)
            if isinstance(conf_matrix, DtlError):
                return conf_matrix

        return MetricsDetails(
            timestamp, metrics_type, epoch, loss, accuracy, precision, recall, f1_score, tpr, fpr, conf_matrix,
        )


class Epoch:
    def __init__(self, epoch_number: int, metrics: List[MetricsDetails]):
        self.epoch_number = epoch_number
        self.metrics = metrics

    def __eq__(self, other: "Epoch"):
        if isinstance(self, other.__class__):
            return self.epoch_number == other.epoch_number and self.metrics == other.metrics
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["epochNo"] = self.epoch_number
        base["metrics"] = list(map(lambda m: m._as_payload(), self.metrics))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Epoch"]:
        """
        Builds an Epoch object from a dictionary,

        :param json: dictionary parsed from json
        :return:
        """

        epoch_number = json.get("epochNo")
        if not isinstance(epoch_number, int):
            return DtlError("An 'epochNo' field is missing")

        metrics = json.get("metrics")
        if metrics is not None:
            metrics = _parse_list(MetricsDetails._from_payload)(metrics)
            if isinstance(metrics, DtlError):
                return metrics
        else:
            metrics = list()

        return Epoch(epoch_number, metrics)


class TrainingState:
    def __init__(
        self,
        id: UUID,
        ontology_id: UUID,
        start_time: str,
        end_time: str,
        status: str,
        user_id: UUID,
        details: str,
        number_of_epochs: int,
        model_type: str,
        epochs: List[Epoch],
        test_metrics: Optional[MetricsDetails] = None,
        endReason: Optional[TrainingEndReason] = None,
        requestedBy: Optional[User] = None,
    ):
        self.start_time = start_time
        self.end_time = end_time
        self.end_reason = endReason
        self.status = status
        self.requested_by = requestedBy
        self.user_id = user_id
        self.details = details
        self.number_of_epochs = number_of_epochs
        self.model_type = model_type
        self.epochs = epochs
        self.test_metrics = test_metrics
        self.ontology_id = ontology_id
        self.id = id

    def __repr__(self):
        return (
            f"(id: {self.id}, ontology_id: {self.ontology_id}, "
            f"start_time: {self.start_time}, status: {self.status}, "
            f"requested_by: {self.requested_by}, user_id: {self.user_id} "
            f"details: {self.details}, number_of_epochs: {self.number_of_epochs},"
            f"model_type: {self.model_type}, epochs: {self.epochs}, test_metrics: {self.test_metrics})"
        )

    def __eq__(self, other: "TrainingState"):
        if isinstance(self, other.__class__):
            return (
                self.id == other.id
                and self.ontology_id == other.ontology_id
                and self.start_time == other.start_time
                and self.end_reason == other.end_reason
                and self.end_time == other.end_time
                and self.status == other.status
                and self.requested_by == other.requested_by
                and self.user_id == other.user_id
                and self.details == other.details
                and self.number_of_epochs == other.number_of_epochs
                and self.model_type == other.model_type
                and self.epochs == other.epochs
                and self.test_metrics == other.test_metrics
            )
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["startTime"] = self.start_time
        base["endTime"] = self.end_time
        base["endReason"] = self.end_reason
        base["requestedBy"] = self.requested_by
        base["userId"] = self.user_id
        base["numberOfEpochs"] = self.number_of_epochs
        base["epochs"] = list(map(lambda e: e._as_payload(), self.epochs))
        base["modelType"] = self.model_type
        base["testMetrics"] = self.test_metrics
        base["ontology_id"] = self.ontology_id
        base["id"] = self.id
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "TrainingState"]:
        """
        Builds a TrainingState object from a dictionary,

        :param json: dictionary parsed from json
        :return:
        """

        end_time = json.get("endTime")
        if not isinstance(end_time, str):
            end_time = None

        end_reason = json.get("endReason")
        if end_reason is not None:
            end_reason = TrainingEndReason._from_payload(end_reason)
            if isinstance(end_reason, DtlError):
                return end_reason

        epochs = json.get("epochs")
        if epochs is not None:
            epochs = _parse_list(Epoch._from_payload)(epochs)
            if isinstance(epochs, DtlError):
                return epochs
        else:
            epochs = list()

        model_type = json.get("modelType")
        if not isinstance(model_type, str):
            return DtlError("A 'modelType' field is missing")

        number_of_epochs = json.get("numberOfEpochs")
        if not isinstance(number_of_epochs, int):
            return DtlError("A 'numberOfEpochs' field is missing")

        ontology_id = json.get("ontologyId")
        if not isinstance(ontology_id, str):
            return DtlError("A 'ontologyId' field is missing")

        model_id = json.get("id")
        if not isinstance(model_id, str):
            return DtlError("An 'id' field is missing")

        start_time = json.get("startTime")
        if not isinstance(start_time, str):
            return DtlError("A 'startTime' field is missing")

        status = json.get("status")
        if not isinstance(status, str):
            return DtlError("A 'status' field is missing")

        test_metrics = json.get("testMetrics")
        if test_metrics is not None:
            test_metrics = MetricsDetails._from_payload(test_metrics)
            if isinstance(test_metrics, DtlError):
                return test_metrics
        else:
            test_metrics = None

        user_id = json.get("userId")
        if not isinstance(user_id, str):
            return DtlError("A 'userId' field is missing")

        details = json.get("details")
        if details is not None:
            if not isinstance(details, str):
                return DtlError("A 'details' field is missing")

        return TrainingState(
            UUID(model_id),
            UUID(ontology_id),
            start_time,
            end_time,
            status,
            UUID(user_id),
            details,
            number_of_epochs,
            model_type,
            epochs,
            test_metrics,
            end_reason,
            None,
        )

    def pd_dataframe_metrics(self):
        """
        Get the tuple object that contains the data frames with the metrics of the specific training.

        :return: Once the tuple is build with `metrics = TrainingState.pd_dataframe_metrics()` the object `metrics` can be accessed:
            - `metrics.description`: data frame with the general information of the training
            - `metrics.validation.metrics[<epoch_number>]`: data frame with precision, recall and f1 for all the classes for that epoch in the validation dataset
            - `metrics.validation.cm[<epoch_number>]`: data frame with confusion matrix for all the classes for that epoch in the validation dataset
            - `metrics.test.metrics`: data frame with precision, recall and f1 for all the classes for that epoch in the test dataset
            - `metrics.test.cm`: data frame with confusion matrix for all the classes for that epoch in the test dataset
            - `metrics.general`: data frame with the general metrics (loss, accuracy, tpr, fpr) for each validation epoch and test set.
        """
        return self._from_training_payload(self)

    @staticmethod
    def _get_description_of_training(training_state: "TrainingState") -> pd.DataFrame:
        """Gets data frame with the description of training.

        :param: training_state: training state object
        :return: data frame with the description of the training (id, ontology id, model type, status, user id, number of epochs,
        start time, end time, end reason and details)
        """
        descrip = {}
        descrip["training_id"] = training_state.id
        descrip["ontology_id"] = training_state.ontology_id
        descrip["model_type"] = training_state.model_type
        descrip["status"] = training_state.status
        descrip["user_id"] = training_state.user_id
        descrip["number_of_epochs"] = training_state.number_of_epochs
        descrip["start_time"] = training_state.start_time
        descrip["end_time"] = training_state.end_time
        descrip["end_reason"] = training_state.end_reason.reason_type if training_state.end_reason is not None else None
        descrip["details"] = training_state.details
        df = pd.DataFrame.from_dict(descrip, orient="index")
        # description columns in the df
        df.columns = ["Description"]

        return df

    @staticmethod
    def _create_validation_confusion_matrices_and_metrics(json_epochs: Epoch) -> tuple:
        """
        Create a tuple with the data frames with the confusion matrices and metrics from the validation set for each epoch.
        Note that the metrics are obtained from the confusion matrix and not the actual training state object.
        :param json_epochs: epochs from training state object (`training_state.epochs`)
        :return:
            1 - tuple with data frames for metrics (3 data frames with precision, recall and f1 for each epoch)
            2 - tuple with data frames for confusion matrices for each epoch
        """
        validation_metrics = {}
        recall_df = pd.DataFrame()
        precision_df = pd.DataFrame()
        f1_df = pd.DataFrame()
        dfs_epochs_cm = dict()
        for epoch_data in json_epochs:
            for epoch_data_metric in epoch_data.metrics:
                if epoch_data_metric.metrics_type == "validation":
                    cm = epoch_data_metric.conf_matrix
                    df = pd.DataFrame(cm.matrix, columns=cm.index, index=cm.index)
                    epoch = "epoch_" + str(epoch_data_metric.epoch)
                    recall = TrainingState._create_recall_df_from_cm_as_df(df, epoch)
                    recall_df = pd.concat([recall_df, recall])
                    precision = TrainingState._create_precision_df_from_cm_as_df(df, epoch)
                    precision_df = pd.concat([precision_df, precision])
                    f1 = TrainingState._create_f1_df_from_recall_precision_as_dfs(precision_df, recall_df, epoch)
                    f1_df = pd.concat([f1_df, f1])
                    df.reset_index(level=0, inplace=True)
                    dfs_epochs_cm[epoch] = df
        validation_metrics["precision"] = precision_df
        validation_metrics["recall"] = recall_df
        validation_metrics["f1"] = f1_df
        for metric in validation_metrics.keys():
            validation_metrics[metric].reset_index(level=0, inplace=True)
        validation_metrics_tuple = TrainingState._from_dict_to_tuple(validation_metrics, "validation_metrics")
        dfs_epochs_cm_tuple = TrainingState._from_dict_to_tuple(dfs_epochs_cm, "confusion_matrix")
        return validation_metrics_tuple, dfs_epochs_cm_tuple

    @staticmethod
    def _create_recall_df_from_cm_as_df(df: pd.DataFrame, indexer: str) -> pd.DataFrame:
        """
        Build the data frame for the recall metrics. The recall is calculated from the confusion matrix.
        :param df: confusion matrix as data frame:
        :param indexer: row description (epochs for validation set and test for test set)
        :return: data frame with recall metrics
        """
        cols = df.columns
        res_df = pd.DataFrame(columns=cols)
        for class_name in cols:
            if df.loc[class_name, :].sum() > 0:
                res_df.loc[indexer, class_name] = df.loc[class_name, class_name] / df.loc[class_name, :].sum()
            else:
                res_df.loc[indexer, class_name] = 0
        return res_df

    @staticmethod
    def _create_precision_df_from_cm_as_df(df: pd.DataFrame, indexer: str) -> pd.DataFrame:
        """
        Build the data frame for the precision metrics. The precision is calculated from the confusion matrix.
        :param df: confusion matrix as data frame:
        :param indexer: row description (epochs for validation set and test for test set)
        :return: data frame with precision metrics
        """
        cols = df.columns
        res_df = pd.DataFrame(columns=cols)
        for class_name in cols:
            if df.loc[:, class_name].sum() > 0:
                res_df.loc[indexer, class_name] = df.loc[class_name, class_name] / df.loc[:, class_name].sum()
            else:
                res_df.loc[indexer, class_name] = 0
        return res_df

    @staticmethod
    def _create_f1_df_from_recall_precision_as_dfs(
        precision: pd.DataFrame, recall: pd.DataFrame, indexer: str
    ) -> Union[pd.DataFrame, DtlError]:
        """
        Build the data frame for the f1 metrics. The f1 is calculated from the precision and recall df.
        :param precision: precision as data frame from confusion matrix:
        :param recall: recall as data frame from confusion matrix:
        :param indexer: row description (epochs for validation set and test for test set)
        :return: data frame with f1 metrics
        """
        if pd.Series(precision.columns).equals(pd.Series(recall.columns)):
            f1_df = pd.DataFrame(columns=precision.columns)
            for col in precision.columns:
                prec = precision.loc[indexer, col]
                rec = recall.loc[indexer, col]
                if prec + rec > 0:
                    f1_df.loc[indexer, col] = 2 * (prec * rec) / (prec + rec)
                else:
                    f1_df.loc[indexer, col] = 0
        else:
            return DtlError("The columns of the Precision - Recall Data Frames doesn't match")
        return f1_df

    @staticmethod
    def _create_test_cm_and_metrics(json_test: MetricsDetails) -> tuple:
        """
        Build data frames with confusion matrix and metrics for the test set.
        :param json_test: json test metrics from training state object (`training_state.test_metrics`)
        :return:
            1 - data frame with confusion matrix
            2 - data frame with test metrics
        """
        cm = json_test.conf_matrix
        df = pd.DataFrame(cm.matrix, columns=cm.index, index=cm.index)
        recall = TrainingState._create_recall_df_from_cm_as_df(df, "temp_index")
        precision = TrainingState._create_precision_df_from_cm_as_df(df, "temp_index")
        f1_df = TrainingState._create_f1_df_from_recall_precision_as_dfs(precision, recall, "temp_index")

        # rename indices
        recall.index = ["recall_test_set"]
        precision.index = ["precision_test_set"]
        f1_df.index = ["f1_test_set"]

        df.reset_index(level=0, inplace=True)

        test_metrics = pd.concat([recall, precision, f1_df])
        test_metrics.reset_index(level=0, inplace=True)
        return df, test_metrics

    @staticmethod
    def _extract_general_metrics(json_file: MetricsDetails) -> pd.DataFrame:
        """Extract the general metrics from validation and test, in a row df.
        :param: json_file: Metrics details. For validation `training_state.epochs[<epoch_num>].metrics[1]` and
        for test set `training_state.test_metrics`
        :return: data frame with the metrics for that validation epoch/test set
        """
        columns = ["loss", "accuracy", "tpr", "fpr", "timestamp"]
        # define the index
        if json_file.metrics_type == "validation":
            ind = "_".join([json_file.metrics_type, "epoch", str(json_file.epoch)])
        elif json_file.metrics_type == "test":
            ind = json_file.metrics_type
        else:
            print("Metrics not in validation or testing...")

        df = pd.DataFrame(index=[ind], columns=columns)
        # build the df with the columns and one index to later concatenate
        for col in columns:
            df.loc[ind, col] = getattr(json_file, col)
        return df

    @staticmethod
    def _build_gen_metrics(training_state: "TrainingState") -> pd.DataFrame:
        """Creates the df with all the gen metrics.
        :param: training_state: training_state object
        :return: df with all the general metrics
        """
        list_dfs = []

        for i in range(len(training_state.epochs)):
            temp_df = TrainingState._extract_general_metrics(training_state.epochs[i].metrics[1])
            list_dfs.append(temp_df)

        if training_state.test_metrics:
            temp_df = TrainingState._extract_general_metrics(training_state.test_metrics)
            list_dfs.append(temp_df)

        if len(list_dfs) > 0:
            df = pd.concat(list_dfs)
            df.reset_index(level=0, inplace=True)

        return df

    @staticmethod
    def _build_list_with_gral_metrics(res_object, metric):
        """
        Builds the precision, recall and f1 general metrics to be added to 'general'
        :param: res_object: dictionary object obtained in previous calculations
        :return: res_object: with metrics added as columns in res_object['general']
        """
        list_metrics = getattr(res_object["validation"]["metrics"], metric).iloc[:, 1:].mean(axis=1).tolist()
        list_metrics.append(
            float(
                res_object["test"]["metrics"]
                .loc[res_object["test"]["metrics"]["index"] == metric + "_test_set"]
                .iloc[:, 1:]
                .mean(axis=1)
            )
        )
        return list_metrics

    @staticmethod
    def _from_dict_to_tuple(dic, name):
        """Build a tuple from a dictionary"""
        return namedtuple(name, dic.keys())(*dic.values())

    @staticmethod
    def _from_training_payload(training_state: "TrainingState") -> tuple:
        """
        Build a tuple with all the metrics in data frames
        :param training_state: training state object
        :return: a tuple containing:
            - `description`: data frame with the general information of the training
            - `validation.metrics[<epoch_number>]`: data frame with precision, recall and f1 for all the classes for that epoch in the validation dataset
            - `validation.cm[<epoch_number>]`: data frame with confusion matrix for all the classes for that epoch in the validation dataset
            - `test.metrics`: data frame with precision, recall and f1 for all the classes for that epoch in the test dataset
            - `test.cm`: data frame with confusion matrix for all the classes for that epoch in the test dataset
            - `general`: data frame with the general metrics (loss, accuracy, tpr, fpr) for each validation epoch and test set.
        """
        res_object = dict()

        res_object["description"] = TrainingState._get_description_of_training(training_state)
        if res_object["description"].loc["end_reason", "Description"] == "Successful":
            res_object["validation"] = dict()
            res_object["test"] = dict()
            (
                res_object["validation"]["metrics"],
                res_object["validation"]["cm"],
            ) = TrainingState._create_validation_confusion_matrices_and_metrics(training_state.epochs)
            (res_object["test"]["cm"], res_object["test"]["metrics"],) = TrainingState._create_test_cm_and_metrics(
                training_state.test_metrics
            )
            res_object["general"] = TrainingState._build_gen_metrics(training_state)
            # add columns precision, recall and f1 to general metrics from previous df's
            for metric in ["precision", "recall", "f1"]:
                res_object["general"][metric] = TrainingState._build_list_with_gral_metrics(res_object, metric)

            # build tuples
            res_object["validation"] = TrainingState._from_dict_to_tuple(res_object["validation"], "validation")
            res_object["test"] = TrainingState._from_dict_to_tuple(res_object["test"], "test")
        else:
            if (
                training_state.status == "Ended"
                and res_object["description"].loc["end_reason", "Description"] != "Successful"
            ):
                message_not_success = f"The current training has ended not being successful. Reason: {res_object['description'].loc['end_reason', 'Description']}"
            else:
                message_not_success = f"At {datetime.datetime.now().strftime('%Y-%m-%d %H:%M')}, the current training has not ended. The status is : {training_state.status}"
            res_object["validation"] = message_not_success
            res_object["general"] = message_not_success
            res_object["test"] = message_not_success
        res_object = TrainingState._from_dict_to_tuple(res_object, "all_metrics")
        return res_object


class Model:
    def __init__(self, id: UUID, states: List[TrainingState]):
        self.id = id
        self.states = states

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["id"] = self.id
        base["states"] = list(map(lambda s: s._as_payload(), self.states))
        return base

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Model"]:
        id = payload.get("trainingId")
        states = payload.get("states")

        if states is not None:
            states = _parse_list(TrainingState._from_payload)(states)
            if isinstance(states, DtlError):
                return states
        else:
            states = list()
        return Model(id=id, states=states)


class DeploymentStatusType(SerializableStringEnum):
    Successful = "Successful"
    Requested = "Requested"
    ModelNotFound = "ModelNotFound"
    InsufficientResources = "InsufficientResources"
    Evicted = "Evicted"
    IllegalEventOccurred = "IllegalEventOccurred"
    ReflexError = "ReflexError"
    EvictionError = "EvictionError"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("deployment status type", s))


def deployment_status_type_from_str(string: str,) -> Union[DtlError, DeploymentStatusType]:
    return SerializableStringEnum.from_str(DeploymentStatusType)(string)


class DeploymentStatus:
    def __init__(self, status_type: DeploymentStatusType, message: Optional[str]):
        self.status_type = status_type
        self.message = message

    def __repr__(self):
        return f"(type: {self.status_type}, details: {self.message})"

    def __eq__(self, other: "DeploymentStatus"):
        if isinstance(self, other.__class__):
            return self.status_type == other.status_type and self.message == other.message
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["_type"] = self.status_type
        base["message"] = self.message
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "DeploymentStatus"]:
        """
            Builds a DeploymentStatus object from a dictionary,

            :param json: dictionary parsed from json
            :return:
            """
        status_type = json.get("_type")
        if status_type is None:
            return DtlError(f"'_type' field is missing")

        status_type = deployment_status_type_from_str(status_type)
        if isinstance(status_type, DtlError):
            return status_type

        message = json.get("message")
        return DeploymentStatus(status_type, message)


class Deployment:
    def __init__(
        self,
        id: UUID,
        ontology_id: UUID,
        model_id: UUID,
        model_type: ModelType,
        status: DeploymentStatus,
        created_at: datetime,
    ):
        self.id = id
        self.ontology_id = ontology_id
        self.model_id = model_id
        self.model_type = model_type
        self.status = status
        self.created_at = created_at

    def __repr__(self):
        return (
            f"Deployment(id: {self.id}, ontology_id: {self.ontology_id}, model_id: {self.model_id},"
            f"model_type: {str(self.model_type)}, status: {str(self.status)}, created_at: {self.created_at})"
        )

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Deployment"]:
        id = map_option(json.get("id"), _parse_uuid)
        if id is None:
            return DtlError(f"'id' field is missing")
        elif isinstance(id, DtlError):
            return id

        ontology_id = map_option(json.get("ontologyId"), _parse_uuid)
        if ontology_id is None:
            return DtlError(f"'ontologyId' field is missing")
        elif isinstance(ontology_id, DtlError):
            return ontology_id

        model_id = map_option(json.get("trainingId"), _parse_uuid)
        if model_id is None:
            return DtlError(f"'trainingId' field is missing")
        elif isinstance(model_id, DtlError):
            return model_id

        model_type = json.get("modelType")
        if model_type is None:
            return DtlError(f"'modelType' field is missing")
        model_type = model_type_from_str(model_type)
        if isinstance(model_type, DtlError):
            return model_type

        status = json.get("status")
        if status is None:
            return DtlError(f"'status' field is missing")

        status = DeploymentStatus._from_payload(status)
        if isinstance(status, DtlError):
            return status

        created_at = json.get("createdAt")
        if created_at is None:
            return DtlError(f"'createdAt' field is missing")
        try:
            created_at = parse(created_at)
        except ValueError:
            return DtlError("'createdAt' could not be parsed as a valid date")

        return Deployment(id, ontology_id, model_id, model_type, status, created_at)


class Annotation:
    def __init__(
        self,
        id: UUID,
        token: str,
        start_char: int,
        end_char: int,
        class_id: UUID,
        timestamp: datetime,
        created_by: UUID,
    ):

        self.id = id
        self.token = token
        self.start_char = start_char
        self.end_char = end_char
        self.class_id = class_id
        self.timestamp = timestamp
        self.created_by = created_by
        """
        One Annotation, as tagged by an NER user
        :params id: id of the annotation
        :params token: the text token that was tagged
        :params start_char: the start position for the token, within the full text of the section
        :params end_char: the end position for the token, within the full text of the section
        :params class_id: id of the class applied with the tag
        :params timestamp: time at which this annotation was created
        :params created_by: id of the user who created this annotation
        """

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "Annotation"]:
        raw_id = json.get("id")
        id = _parse_uuid(raw_id)
        token = json.get("token")
        start_char = json.get("startChar")
        end_char = json.get("endChar")
        raw_class_id = json.get("classId")
        class_id = _parse_uuid(raw_class_id)
        timestamp = json.get("timestamp")
        raw_created_by = json.get("createdBy")
        created_by = _parse_uuid(raw_created_by)
        return Annotation(id, token, start_char, end_char, class_id, timestamp, created_by)

    @staticmethod
    def from_array(items: List[dict]) -> Union[DtlError, List["Annotation"]]:
        results = []
        for item in items:
            parsed = Annotation.from_payload(item)
            if isinstance(parsed, DtlError):
                return parsed
            results.append(parsed)
        return results

    def __repr__(self):
        return f"{self.__class__.__name__}(id: {self.id}, token: {self.token}, start_char: {self.start_char}, end_char: {self.end_char}, class_id: {self.class_id}, timestamp: {self.timestamp}, created_by: {self.created_by})"


class Section:
    def __init__(self, datastore_id: UUID, value: str, annotations: List[Annotation]):
        self.datastore_id = datastore_id
        self.value = value
        self.annotations = annotations
        """
        A section of unstructured data that has been annotated in the NER flow
        :params datastore_id: id of the datastore the section was imported from
        :params section_full_text: the full text of the section
        :params annotations: a list of Annotation objects, each with relevant statistics
        """

    def __repr__(self):
        return f"{self.__class__.__name__}(datastore_id: {self.datastore_id}\n value: {self.value}\n annotations: {self.annotations!r})"

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "Section"]:
        value = json.get("value")
        datastore_id = json.get("datastoreId")
        annotations_raw = json.get("annotations")
        if not isinstance(annotations_raw, List) and annotations_raw is not None:
            return DtlError(f"annotations of {datastore_id} should be a list")
        else:
            annotations = []
            if isinstance(annotations_raw, List):
                annotations = _parse_list(Annotation.from_payload)(annotations_raw)
            return Section(datastore_id, value, annotations)


class NerTag:
    def __init__(self, start_token: int, end_token: int, tag: UUID):
        self.start_token = start_token
        self.end_token = end_token
        self.tag = tag

    def __eq__(self, other):
        if isinstance(other, NerTag):
            return self.start_token == other.start_token and self.end_token == other.end_token and self.tag == other.tag
        else:
            return False

    def __repr__(self):
        return f"{self.__class__.__name__}(start_token: {self.start_token}\n end_token: {self.end_token}\n tag: {self.tag!r})"

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "NerTag"]:
        start_token = json.get("startToken")
        if not isinstance(start_token, int):
            return DtlError("A 'startToken' field is missing")

        end_token = json.get("endToken")
        if not isinstance(end_token, int):
            return DtlError("A 'endToken' field is missing")

        raw_tag = json.get("tag")
        if not isinstance(raw_tag, str):
            return DtlError("A 'tag' field is missing")
        tag = _parse_uuid(raw_tag)
        if not isinstance(tag, UUID):
            return DtlError("A 'tag' should be valid uuid")

        return NerTag(start_token, end_token, tag)

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = dict()
        base["startToken"] = self.start_token
        base["endToken"] = self.end_token
        base["tag"] = str(self.tag)
        return base


class Token:
    def __init__(self, value: str, type: str):
        self.value = value
        self.type = type


class OriginalSection:
    def __init__(
        self,
        section_id: UUID,
        datastore_id: UUID,
        source_path: List[str],
        tokens: List[str],
        token_types: List[str],
        rejected: bool,
        tags: List[List[NerTag]],
    ):
        self.section_id = section_id
        self.datastore_id = datastore_id
        self.source_path = source_path
        self.tokens = tokens
        self.token_types = token_types
        self.rejected = rejected
        self.tags = tags

    def __repr__(self):
        return (
            f"Section(section_id: {self.section_id}, "
            f"datastore_id: {self.datastore_id}, "
            f"source_path: {self.source_path},"
            f"tokens: {self.tokens},"
            f"token_types: {self.token_types},"
            f"rejected: {self.rejected},"
            f"tags: {self.tags},"
            f")"
        )

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "OriginalSection"]:

        raw_section_id = json.get("sectionId")
        if not isinstance(raw_section_id, str):
            return DtlError("A 'sectionId' field is missing")
        section_id = _parse_uuid(raw_section_id)
        if not isinstance(section_id, UUID):
            return DtlError("A 'sectionId' should be valid uuid")

        raw_datastore_id = json.get("datastoreId")
        if not isinstance(raw_datastore_id, str):
            return DtlError("A 'datastoreId' field is missing")
        datastore_id = _parse_uuid(raw_datastore_id)
        if not isinstance(datastore_id, UUID):
            return DtlError("A 'datastoreId' should be valid uuid")

        source_path = json.get("sourcePath")
        if not isinstance(source_path, List):
            return DtlError("A 'sourcePath' field is missing")

        tokens = json.get("tokens")
        if not isinstance(tokens, List):
            return DtlError("A 'tokens' field is missing")

        token_types = json.get("tokenTypes")
        if not isinstance(token_types, List):
            return DtlError("A 'tokenTypes' field is missing")

        rejected = json.get("rejected")
        if not isinstance(rejected, bool):
            return DtlError("A 'tokenTypes' field is missing")

        raw_tags = json.get("tags")
        if not isinstance(raw_tags, List):
            return DtlError("A 'tags' field is missing")
        all_parsed_tags = []
        for annotation in raw_tags:
            if not isinstance(annotation, List):
                return DtlError("A 'tags' field is missing")
            annotation_tags = _parse_list(NerTag.from_payload)(annotation)
            if isinstance(annotation_tags, DtlError):
                return annotation_tags
            all_parsed_tags.append(annotation_tags)

        return OriginalSection(section_id, datastore_id, source_path, tokens, token_types, rejected, all_parsed_tags,)

    def replace_tag_by(
        self, test: Callable[[NerTag], bool], generate_tokens: Callable[[], List[Token]]
    ) -> List["SynthesizedSection"]:
        """
        Create new sections
        replace tag when 'test' is true using 'generate_tokens' method to replace original tokens in the section
        Returned list of NewSections to be created as augmented data.
        :params test - function to determine replacement
        :params generate_tokens - function to generate new tokens and token types
        """

        def generator():
            return [generate_tokens()]

        return self.replace_tag_by_many(test, generator)

    def replace_tag_by_many(
        self, test: Callable[[NerTag], bool], generate_tokens: Callable[[], List[List[Token]]],
    ) -> List["SynthesizedSection"]:
        """
        Create new sections
        replace tag when 'test' is true using 'generate_tokens' method to replace original tokens in the section
        Returned list of NewSections to be created as augmented data.
        Every item in generate_tokens list will generate a new section in the output.
        :params test - function to determine replacement
        :params generate_tokens - function to generate new tokens and token types
        """

        def replace_tag(
            section: OriginalSection, tag: NerTag, all_tags: Iterator[NerTag]
        ) -> List["SynthesizedSection"]:
            before_tokens = section.tokens[: tag.start_token]
            after_tokens = section.tokens[tag.end_token :]

            before_token_types = section.token_types[: tag.start_token]
            after_token_types = section.token_types[tag.end_token :]

            new_tokens_definition = generate_tokens()

            tag_sections = []
            for token_definition in new_tokens_definition:
                new_tokens = [t.value for t in token_definition]
                new_token_types = [t.type for t in token_definition]
                tokens = before_tokens + new_tokens + after_tokens
                token_types = before_token_types + new_token_types + after_token_types

                before_tags = []
                after_tags = []

                for t in all_tags:
                    if t.start_token < tag.start_token:
                        before_tags.append(t)
                    if t.start_token > tag.start_token:
                        after_tags.append(t)

                old_length = tag.end_token - tag.start_token
                new_length = len(new_tokens)

                length_diff = new_length - old_length

                def adjust_tags_positions(t: NerTag) -> NerTag:
                    return NerTag(t.start_token + length_diff, t.end_token + length_diff, t.tag)

                after_tags = [adjust_tags_positions(t) for t in after_tags]
                tag_sections.append(
                    SynthesizedSection(
                        None,
                        section.section_id,
                        section.datastore_id,
                        section.source_path,
                        tokens,
                        token_types,
                        section.rejected,
                        before_tags
                        + [NerTag(tag.start_token, tag.start_token + len(new_tokens), tag.tag,)]
                        + after_tags,
                    )
                )
                return tag_sections

        tags_flatten = list(itertools.chain.from_iterable(self.tags))
        to_flatten = [replace_tag(self, tag, tags_flatten) for tag in tags_flatten if test(tag)]
        return list(itertools.chain.from_iterable(to_flatten))

    def __eq__(self, other):
        if isinstance(other, OriginalSection):
            return (
                self.section_id == other.section_id
                and self.datastore_id == other.datastore_id
                and self.source_path == other.source_path
                and self.tokens == other.tokens
                and self.token_types == other.token_types
                and self.tags == other.tags
            )

        else:
            return False


class SynthesizedSection:
    def __init__(
        self,
        section_id: Optional[UUID],
        original_id: UUID,
        datastore_id: UUID,
        source_path: List[str],
        tokens: List[str],
        token_types: List[str],
        rejected: bool,
        tags: List[NerTag],
    ):
        self.section_id = section_id
        self.original_id = original_id
        self.datastore_id = datastore_id
        self.source_path = source_path
        self.tokens = tokens
        self.token_types = token_types
        self.rejected = rejected
        self.tags = tags

    def __repr__(self):
        return (
            f"Section(section_id: {self.section_id}, "
            f"datastore_id: {self.datastore_id}, "
            f"source_path: {self.source_path},"
            f"tokens: {self.tokens},"
            f"token_types: {self.token_types},"
            f"rejected: {self.rejected},"
            f"tags: {self.tags},"
            f"original_id: {self.original_id}"
            f")"
        )

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "SynthesizedSection"]:

        raw_section_id = json.get("sectionId")
        if not isinstance(raw_section_id, str):
            return DtlError("A 'sectionId' field is missing")
        section_id = _parse_uuid(raw_section_id)
        if not isinstance(section_id, UUID):
            return DtlError("A 'sectionId' should be valid uuid")

        raw_original_id = json.get("originalId")
        if isinstance(raw_original_id, str):
            original_id = _parse_uuid(raw_original_id)
            if isinstance(original_id, DtlError):
                return original_id

        raw_datastore_id = json.get("datastoreId")
        if not isinstance(raw_datastore_id, str):
            return DtlError("A 'datastoreId' field is missing")
        datastore_id = _parse_uuid(raw_datastore_id)
        if not isinstance(datastore_id, UUID):
            return DtlError("A 'datastoreId' should be valid uuid")

        source_path = json.get("sourcePath")
        if not isinstance(source_path, List):
            return DtlError("A 'sourcePath' field is missing")

        tokens = json.get("tokens")
        if not isinstance(tokens, List):
            return DtlError("A 'tokens' field is missing")

        token_types = json.get("tokenTypes")
        if not isinstance(token_types, List):
            return DtlError("A 'tokenTypes' field is missing")

        rejected = json.get("rejected")
        if not isinstance(rejected, bool):
            return DtlError("A 'tokenTypes' field is missing")

        raw_tags = json.get("tags")
        if not isinstance(raw_tags, List):
            return DtlError("A 'tags' field is missing")
        tags = _parse_list(NerTag.from_payload)(raw_tags)

        return SynthesizedSection(
            section_id, original_id, datastore_id, source_path, tokens, token_types, rejected, tags,
        )

    def __eq__(self, other):
        if isinstance(other, SynthesizedSection):
            return (
                self.section_id == other.section_id
                and self.original_id == other.original_id
                and self.datastore_id == other.datastore_id
                and self.source_path == other.source_path
                and self.tokens == other.tokens
                and self.token_types == other.token_types
                and self.tags == other.tags
            )

        else:
            return False

    def _as_payload(self):
        return {
            "datastoreId": str(self.datastore_id),
            "sourcePath": self.source_path,
            "tokens": self.tokens,
            "tokenTypes": self.token_types,
            "rejected": self.rejected,
            "tags": [tag._as_payload() for tag in self.tags],
            "originalId": str(self.original_id),
        }


class SynthesisStats:
    def __init__(
        self,
        generated_sections: int,
        processed_sections: int,
        avg_number_generated_sections: float,
        std_number_generated_section: float,
    ):
        """
        :param generated_sections: total number of generated sections
        :param processed_sections: total number that were used as input
        :param avg_number_generated_sections: average number of sections generated by original section
        :param std_number_generated_section: standard deviation of the number of generated sections
        """
        self.generated_sections = generated_sections
        self.processed_sections = processed_sections
        self.avg_number_generated_sections = avg_number_generated_sections
        self.std_number_generated_section = std_number_generated_section

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(generated_sections: {self.generated_sections},\n"
            f"processed_sections: {self.processed_sections})"
        )

    def __eq__(self, other):
        if isinstance(other, SynthesisStats):
            return (
                self.generated_sections == other.generated_sections
                and self.processed_sections == other.processed_sections
                and self.avg_number_generated_sections == other.avg_number_generated_sections
                and self.std_number_generated_section == other.std_number_generated_section
            )



from typing import Dict, Union

from datalogue.errors import DtlError
from datalogue.version import __version__ as dtl_version


class Version:
    """
    Version object containing version number of SDK, platform, and the platform's services'
    """

    def __init__(self, sdk: str, platform: str, services: Dict[str, str]):
        """
            :param sdk: sdk version
            :param platform: Datalogue's platform version
            :param services: Datalogue's list of services
        """
        self.sdk = sdk
        self.platform = platform
        self.services = services

    def __eq__(self, other):
        return (self.sdk == other.sdk) and (self.platform == other.platform) and (self.services == other.services)

    def __repr__(self):
        service_string = ""
        for service, version in self.services.items():
            service_string += "    " + service + ": " + version + "\n"
        return (
            "DTLVersion("
            f"\n  sdk: {self.sdk}, "
            f"\n  platform: {self.platform}, "
            f"\n  services:\n{service_string})"
        )

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, "Version"]:
        """
        Convert to Datalogue's version object from JSON payload received from Datalogue's backend
        :param payload: The JSON payload
        :return: Datalogue's version object
        """
        platform_ver = payload.get("platform")
        if platform_ver is None or not isinstance(platform_ver, str):
            return DtlError("'platform_ver' is missing or not a string")
        services_ver = payload.get("services")
        if services_ver is None or not isinstance(services_ver, dict):
            return DtlError("'services_ver' is missing or not a dictionary")
        return Version(dtl_version, platform_ver, services_ver)



from datetime import datetime
from uuid import UUID
from typing import Union, Optional, List

from dateutil.parser import parse

from datalogue.dtl_utils import SerializableStringEnum
from datalogue.dtl_utils import _parse_list, _parse_uuid
from datalogue.errors import _enum_parse_error, DtlError


class JobStatus(SerializableStringEnum):
    Scheduled = "Scheduled"
    Defined = "Defined"
    Running = "Running"
    Succeeded = "Succeeded"
    Failed = "Failed"
    Unknown = "Unknown"
    Cancelled = "Cancelled"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("job status", s))

    @staticmethod
    def job_status_from_str(string: str) -> Union[DtlError, "JobStatus"]:
        return SerializableStringEnum.from_str(JobStatus)(string)


class Job:
    def __init__(
        self,
        id: UUID,
        pipeline_id: Optional[UUID],
        run_at: datetime,
        status: JobStatus,
        remaining_time_millis: Optional[int],
        percentage_progress: Optional[float],
        created_by: UUID,
        errors: Optional[str],
        started_at: Optional[datetime],
        ended_at: Optional[datetime],
        average_records_speed: float,
        current_records_speed: float,
        estimated_processed_records_count: int,
        average_bytes_speed: Optional[float],
        current_bytes_speed: Optional[float],
        estimated_processed_bytes_count: Optional[int],
        total_source_records: Optional[int],
        total_source_bytes: Optional[int],
        wait_on: List[UUID] = [],
        abort_if_overlong: bool = True,
        frequency: Optional[str] = None,
        parent_job_id: Optional[UUID] = None,
    ):
        self.pipeline_id = pipeline_id
        self.status = status
        self.run_at = run_at
        self.id = id
        self.remaining_time_millis = remaining_time_millis
        self.percentage_progress = percentage_progress
        self.created_by = created_by
        self.errors = errors
        self.ended_at = ended_at
        self.started_at = started_at
        self.wait_on = wait_on
        self.abort_if_overlong = abort_if_overlong
        self.frequency = frequency
        self.parent_job_id = parent_job_id

        self.average_records_speed = average_records_speed
        self.current_records_speed = current_records_speed
        self.estimated_processed_records_count = estimated_processed_records_count
        self.average_bytes_speed = average_bytes_speed
        self.current_bytes_speed = current_bytes_speed
        self.estimated_processed_bytes_count = estimated_processed_bytes_count
        self.total_source_records = total_source_records
        self.total_source_bytes = total_source_bytes

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"id: {self.id!r}, "
            f"pipeline_id: {self.pipeline_id!r}, "
            f"status: {self.status!r}, "
            f"run_at: {self.run_at!r}, "
            f"wait_on: {self.wait_on!r}, "
            f"created_by: {self.created_by!r}, "
            f"remaining_time_millis: {self.remaining_time_millis!r}, "
            f"percent_progress: {self.percentage_progress!r}, "
            f"errors: {self.errors!r}, "
            f"average_records_speed: {self.average_records_speed!r}, "
            f"current_records_speed: {self.current_records_speed!r}, "
            f"estimated_processed_records_count: {self.estimated_processed_records_count!r}, "
            f"average_bytes_speed: {self.average_bytes_speed!r}, "
            f"current_bytes_speed: {self.current_bytes_speed!r}, "
            f"estimated_processed_bytes_count: {self.estimated_processed_bytes_count!r}, "
            f"total_source_records: {self.total_source_records!r}, "
            f"total_source_bytes: {self.total_source_bytes!r}, "
            f"ended_at: {self.ended_at!r}, "
            f"started_at: {self.started_at!r}, "
            f"abort_if_overlong: {self.abort_if_overlong}, "
            f"frequency: {self.frequency}, "
            f"parent_job_id: {self.parent_job_id})"
        )


def _job_from_payload(json: dict) -> Union[DtlError, Job]:
    job_id = json.get("jobId")
    if job_id is None:
        return DtlError("Job object should have a 'jobId' property")
    else:
        try:
            job_id = UUID(job_id)
        except ValueError:
            return DtlError("'jobId' field was not a proper uuid")

    pipeline_id = json.get("streamId")
    if pipeline_id is not None:
        try:
            pipeline_id = UUID(pipeline_id)
        except ValueError:
            return DtlError("'streamId' field was not a proper uuid")

    run_at = json.get("runDate")
    if run_at is None:
        return DtlError("Job object should have a 'runDate' property")
    else:
        try:
            run_at = parse(run_at)
        except ValueError:
            return DtlError("The 'runDate' could not be parsed as a valid date")

    status = json.get("status")
    if status is None:
        return DtlError("Job object should have a 'status' property")
    else:
        status = JobStatus.job_status_from_str(status)
        if isinstance(status, DtlError):
            return status

    remaining_time_millis = json.get("remainingTimeInMillis")
    if remaining_time_millis is not None:
        try:
            remaining_time_millis = int(remaining_time_millis)
        except ValueError:
            return DtlError("'remainingTimeInMillis' was not a proper int")

    percentage_progress = json.get("percentage")
    if percentage_progress is None:
        percentage_progress = None
    else:
        try:
            percentage_progress = float(percentage_progress)
        except ValueError:
            return DtlError("'percentage' was not a proper number")

    created_by = json.get("createdBy")
    if created_by is None:
        return DtlError("Job object should have a 'createdBy' property")
    else:
        try:
            created_by = UUID(created_by)
        except ValueError:
            return DtlError("'createdBy' was not a proper uuid")

    # details is optional
    errors = json.get("details")

    # ended_at is optional
    ended_at = json.get("endDate")
    if ended_at is not None:
        try:
            ended_at = parse(ended_at)
        except ValueError:
            return DtlError("The 'endDate' could not be parsed as a valid date")

    # started_at is optional
    started_at = json.get("startDate")
    if started_at is not None:
        try:
            started_at = parse(started_at)
        except ValueError:
            return DtlError("The 'startDate' could not be parsed as a valid date")

    wait_on = json.get("waitOn")
    if wait_on is not None:
        wait_on = _parse_list(_parse_uuid)(wait_on)
        if isinstance(wait_on, DtlError):
            return wait_on
    else:
        return DtlError("Job object should have a 'waitOn' property")

    frequency = None
    if json.get("intervalInSec") is not None:
        # TODO rename in scout so we have the same naming in both
        frequency = str(json.get("intervalInSec"))

    parent_job_id = None
    if json.get("parentJobId") is not None:
        parent_job_id = json.get("parentJobId")

    abort_if_overlong = json.get("abortIfOverlong")

    average_records_speed = json.get("averageRecordsSpeed")
    if average_records_speed is None:
        return DtlError("Job object should have a 'averageRecordsSpeed' property")
    else:
        try:
            average_records_speed = float(average_records_speed)
        except ValueError:
            return DtlError("'averageRecordsSpeed' was not a proper number")

    current_records_speed = json.get("currentRecordsSpeed")
    if current_records_speed is None:
        return DtlError("Job object should have a 'currentRecordsSpeed' property")
    else:
        try:
            current_records_speed = float(current_records_speed)
        except ValueError:
            return DtlError("'currentRecordsSpeed' was not a proper number")

    estimated_processed_records_count = json.get("estimatedProcessedRecordsCount")
    if estimated_processed_records_count is None:
        return DtlError("Job object should have a 'estimatedProcessedRecordsCount' property")
    else:
        try:
            estimated_processed_records_count = int(estimated_processed_records_count)
        except ValueError:
            return DtlError("'estimatedProcessedRecordsCount' was not a proper int")

    average_bytes_speed = json.get("averageBytesSpeed")
    if average_bytes_speed is not None:
        try:
            average_bytes_speed = float(average_bytes_speed)
        except ValueError:
            return DtlError("'averageBytesSpeed' was not a proper number")

    current_bytes_speed = json.get("currentBytesSpeed")
    if current_bytes_speed is not None:
        try:
            current_bytes_speed = float(current_bytes_speed)
        except ValueError:
            return DtlError("'currentBytesSpeed' was not a proper number")

    estimated_processed_bytes_count = json.get("estimatedProcessedBytesCount")
    if estimated_processed_bytes_count is not None:
        try:
            estimated_processed_bytes_count = int(estimated_processed_bytes_count)
        except ValueError:
            return DtlError("'estimatedProcessedBytesCount' was not a proper int")

    total_source_records = json.get("totalSourceRecords")
    if total_source_records is not None:
        try:
            total_source_records = int(total_source_records)
        except ValueError:
            return DtlError("'totalSourceRecords' was not a proper int")

    total_source_bytes = json.get("totalSourceBytes")
    if total_source_bytes is not None:
        try:
            total_source_bytes = int(total_source_bytes)
        except ValueError:
            return DtlError("'totalSourceBytes' was not a proper int")

    job = Job(
        id=job_id,
        pipeline_id=pipeline_id,
        run_at=run_at,
        status=status,
        remaining_time_millis=remaining_time_millis,
        percentage_progress=percentage_progress,
        created_by=created_by,
        errors=errors,
        started_at=started_at,
        ended_at=ended_at,
        average_records_speed=average_records_speed,
        current_records_speed=current_records_speed,
        estimated_processed_records_count=estimated_processed_records_count,
        average_bytes_speed=average_bytes_speed,
        current_bytes_speed=current_bytes_speed,
        estimated_processed_bytes_count=estimated_processed_bytes_count,
        total_source_records=total_source_records,
        total_source_bytes=total_source_bytes,
        wait_on=wait_on,
        abort_if_overlong=abort_if_overlong,
        frequency=frequency,
        parent_job_id=parent_job_id,
    )

    return job




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: dtl-python-sdk:datalogue

from setuptools import setup

setup(**{
    'name': 'datalogue',
    'version': '1.10.0-RC.797018975',
    'author': 'Nicolas Joseph',
    'author_email': 'nic@datalogue.io',
    'license': """
        Copyright 2021 Datalogue, Inc.

        This Datalogue SDK is licensed solely pursuant to the terms of the Master Software License executed between you as Licensee and Datalogue, Inc.

        All rights reserved.
        """,
    'description': 'SDK to interact with the datalogue platform',
    'long_description': '',
    'long_description_content_type': 'text/markdown',
    'url': 'https://github.com/datalogue/platform',
    'classifiers': [
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
    ],
    'python_requires': '>=3.6',
    'setup_requires': [
        'pytest-runner',
    ],
    'tests_require': [
        'pytest>=3.6.3',
        'pytest-cov>=2.6.0',
    ],
    'package_dir': {
        '': 'src',
    },
    'packages': [
        'datalogue',
        'datalogue.anduin',
        'datalogue.anduin.models',
        'datalogue.auth',
        'datalogue.clients',
        'datalogue.models',
        'datalogue.models.kinesis',
        'datalogue.models.transformations',
    ],
    'package_data': {
    },
    'install_requires': [
        'requests',
        'python-dateutil',
        'validators',
        'pytest>=3.6.3',
        'numpy>=1.19.4',
        'pyyaml',
        'pyarrow',
        'pandas>=1.1.5',
        'pbkdf2',
    ],
})



import abc
from typing import Optional, Union
from datalogue.models.datastore import CredentialType, ChangeDataSource
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models.s3encryption import Encryption
from datalogue.dtl_utils import SerializableStringEnum
from uuid import UUID


class CredentialsReference:
    def __init__(self, ref_id: UUID, name: Optional[str], ref_type: CredentialType):
        self.id = ref_id
        self.name = name
        self.type = ref_type

    def __eq__(self, other: "CredentialsReference"):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.name == other.name and self.type == other.type
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}(id: {self.id}, name: {self.name!r}, type: {self.type!r})"


def _credentials_ref_from_payload(json: dict) -> Union[DtlError, CredentialsReference]:
    ref_id = json.get("id")
    if not isinstance(ref_id, str):
        return DtlError("A credentials reference needs an 'id' field")
    else:
        ref_id = UUID(ref_id)

    ref_type = json.get("type")
    if not isinstance(ref_type, str):
        return DtlError("A credentials reference needs a '_type' field")
    else:
        ref_type = CredentialType.credential_type_from_str(ref_type)

    name = json.get("name")
    return CredentialsReference(ref_id, name, ref_type)


class Credentials(abc.ABC):
    type_field = "type"

    def __init__(self, definition_type: CredentialType):
        self.type = definition_type

    def _base_payload(self) -> dict:
        return dict([(Credentials.type_field, self.type.value)])

    def __eq__(self, other: "Credentials"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """


class S3Credentials(Credentials):
    type_str = CredentialType.S3

    def __init__(
        self,
        region: str,
        connection_bucket: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        endpoint: Optional[str] = None,
        encryption: Optional[Encryption] = None,
    ):
        Credentials.__init__(self, S3Credentials.type_str)
        self.region = region
        is_client_id_set = client_id is not None
        is_client_secret_set = client_secret is not None
        if is_client_id_set != is_client_secret_set:
            raise ValueError("Client id and client secret must but set, or both can't be set")
        if is_client_id_set and is_client_secret_set:
            self.client_id = client_id
            self.client_secret = client_secret
        self.connection_bucket = connection_bucket
        self.endpoint = endpoint
        self.encryption = encryption

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(region: {self.region!r}, connection_bucket: {self.connection_bucket!r}, "
            f"client_id: {self.client_id!r}, client_secret: "
            f"****, endpoint: {self.endpoint!r}) "
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        if self.connection_bucket is not None:
            base["connectionBucket"] = self.connection_bucket
        if self.client_id is not None and self.client_secret is not None:
            base["clientId"] = self.client_id
            base["clientSecret"] = self.client_secret
        if self.endpoint is not None:
            base["endpoint"] = self.endpoint
        if self.encryption is not None:
            base["encryption"] = self.encryption._as_payload()
        return base


class S3ACredentials(S3Credentials):
    type_str = CredentialType.Hadoop

    def __init__(self, region: str, client_id: str, client_secret: str, endpoint: str):
        S3Credentials.__init__(self, region, client_id, client_secret, endpoint)
        self.type = S3ACredentials.type_str


class AmazonVendorCentral(Credentials):
    type_str = CredentialType.AmazonVendorCentral

    def __init__(self, client_id: str, client_secret: str, mws_auth_token: str, seller_id: str):
        Credentials.__init__(self, AmazonVendorCentral.type_str)
        self.client_id = client_id
        self.client_secret = client_secret
        self.mws_auth_token = mws_auth_token
        self.seller_id = seller_id

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(client_id: ***, client_secret: ***, "
            f"mws_auth_token: ***, seller_id: {self.seller_id})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        base["mwsAuthToken"] = self.mws_auth_token
        base["sellerId"] = self.seller_id
        return base


class AzureBlobType(SerializableStringEnum):
    Block = "Block"
    Append = "Append"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, "AzureBlobType"]:
        return SerializableStringEnum.from_str(AzureBlobType)(json)


class AzureEndpointProtocol(SerializableStringEnum):
    Https = "Https"
    Http = "Http"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, "AzureBlobType"]:
        return SerializableStringEnum.from_str(AzureEndpointProtocol)(json)


class AzureCredentials(Credentials):
    type_str = CredentialType.Azure

    def __init__(
        self, account_name: str, account_key: str, endpoint_protocol: AzureEndpointProtocol, blob_type: AzureBlobType,
    ):
        Credentials.__init__(self, AzureCredentials.type_str)
        self.account_name = account_name
        self.account_key = account_key
        self.endpoint_protocol = endpoint_protocol
        self.blob_type = blob_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(account_name: {self.account_name!r},"
            f" endpoint_protocol: {self.endpoint_protocol!r}, blob_type: {self.blob_type!r}, account_key: ****)"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["accountName"] = self.account_name
        base["accountKey"] = self.account_key
        base["endpointProtocol"] = self.endpoint_protocol.value
        base["blobType"] = self.blob_type.value
        return base


class HadoopAzureCredentials(Credentials):
    type_str = CredentialType.HadoopAzure

    def __init__(self, account_name: str, account_key: str):
        Credentials.__init__(self, HadoopAzureCredentials.type_str)
        self.account_name = account_name
        self.account_key = account_key

    def __repr__(self):
        return f"{self.__class__.__name__}(account_name: {self.account_name!r}," f" account_key: ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["accountName"] = self.account_name
        base["accountKey"] = self.account_key
        return base


class AzureSASCredentials(Credentials):
    type_str = CredentialType.Azure

    def __init__(self, sas_token: str, endpoint: str):
        Credentials.__init__(self, AzureSASCredentials.type_str)
        self.sas_token = sas_token
        self.endpoint = endpoint

    def __repr__(self):
        return f"{self.__class__.__name__}(sas_token: ****," f" endpoint: {self.endpoint!r})"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["sas_token"] = self.sas_token
        base["endpoint"] = self.endpoint
        return base


class GCS(Credentials):
    type_str = CredentialType.GCS

    def __init__(self, client_email: str, private_key: str):
        Credentials.__init__(self, GCS.type_str)
        self.client_email = client_email
        self.private_key = private_key

    def __repr__(self):
        return f"{self.__class__.__name__}(client_email: {self.client_email!r}, private_key: ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["clientEmail"] = self.client_email
        base["privateKey"] = self.private_key
        return base


class JdbcCredentials(Credentials):
    type_str = CredentialType.JDBC

    def __init__(self, url: str, user: str, password: str):
        Credentials.__init__(self, JdbcCredentials.type_str)
        self.url = url
        self.user = user
        self.password = password

    def __repr__(self):
        return f"{self.__class__.__name__}(url: {self.url!r}, user: {self.user!r}, password: ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["user"] = self.user
        base["password"] = self.password
        return base


class MongoCredentials(Credentials):
    type_str = CredentialType.Mongo

    def __init__(self, url: str, database: str, user: str, password: str):
        Credentials.__init__(self, MongoCredentials.type_str)
        self.url = url
        self.database = database
        self.user = user
        self.password = password

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(url: {self.url!r}, database: {self.database!r}, user: {self.user!r}, "
            f"password: ****) "
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["database"] = self.database
        base["user"] = self.user
        base["password"] = self.password
        return base


class SocrataCredentials(Credentials):
    type_str = CredentialType.Socrata

    def __init__(self, domain: str, token: str):
        Credentials.__init__(self, SocrataCredentials.type_str)
        self.domain = domain
        self.token = token

    def __repr__(self):
        return f"{self.__class__.__name__}(domain: {self.domain!r}, token: ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["domain"] = self.domain
        base["token"] = self.token
        return base


class KinesisCredentials(Credentials):
    type_str = CredentialType.Kinesis

    def __init__(self, region: str, client_id: str, client_secret: str):
        Credentials.__init__(self, KinesisCredentials.type_str)
        self.region = region
        self.client_id = client_id
        self.client_secret = client_secret

    def __repr__(self):
        return f"{self.__class__.__name__}(region: {self.region!r}, client_id: {self.client_id!r}, client_secret: ****)"

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        return base


class KerberosCredentials(Credentials):
    type_str = CredentialType.Kerberos

    def __init__(self, principal: str, password: str, with_ssl: bool = True):
        Credentials.__init__(self, KerberosCredentials.type_str)
        self.principal = principal
        self.password = password
        self.with_ssl = with_ssl

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(principal: {self.principal!r}, "
            f"password: ****, "
            f"with_ssl: {self.with_ssl})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["principal"] = self.principal
        base["password"] = self.password
        base["withSsl"] = self.with_ssl
        base["type"] = "Kerberos"
        return base


class JaasString(Credentials):
    type_str = CredentialType.JaasString

    def __init__(self, sasl_jaas_config: str, with_ssl: bool = True):
        Credentials.__init__(self, KerberosCredentials.type_str)
        self.sasl_jaas_config = sasl_jaas_config
        self.with_ssl = with_ssl

    def __repr__(self):
        return (
            f"{self.__class__.__name__}" f"(saslJaasConfig: {self.sasl_jaas_config!r}, " f"with_ssl: {self.with_ssl})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["saslJaasConfig"] = self.sasl_jaas_config
        base["withSsl"] = self.with_ssl
        base["type"] = "JaasString"
        return base


class KafkaCredentials(Credentials):
    type_str = CredentialType.Kafka

    def __init__(
        self,
        bootstrap_server: str,
        registry_url: Optional[str] = None,
        auth: Union[KerberosCredentials, JaasString, None] = None,
        change_data_source: Optional[ChangeDataSource] = None,
    ):
        Credentials.__init__(self, KafkaCredentials.type_str)
        self.bootstrap_server = bootstrap_server
        self.registry_url = registry_url
        self.auth = auth
        self.change_data_source = change_data_source

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(bootstrap_server: {self.bootstrap_server!r}, "
            f"registry_url: {self.registry_url}, "
            f"auth: {self.auth}, "
            f"change_data_source: {self.change_data_source})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["bootstrapServer"] = self.bootstrap_server
        if self.registry_url is not None:
            base["registryUrl"] = self.registry_url
        if isinstance(self.auth, KerberosCredentials):
            base["auth"] = KerberosCredentials._as_payload(self.auth)
        if isinstance(self.auth, JaasString):
            base["auth"] = JaasString._as_payload(self.auth)
        if self.change_data_source is not None:
            base["changeDataSource"] = self.change_data_source
        return base


class ResourceType(SerializableStringEnum):
    """
    Class that handles all ResourceTypes
    """

    File = "File"
    Directory = "Directory"
    Unknown = "Unknown"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("Resource type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, "ResourceType"]:
        return SerializableStringEnum.from_str(ResourceType)(string)


class SFTPCredentials(Credentials):
    type_str = CredentialType.SFTP

    def __init__(
        self,
        address: str,
        port: int,
        user_name: str,
        private_key: str,
        encryption_key: str = "",
        target_resource_path: str = "",
        target_resource_type: Optional[ResourceType] = None,
    ):
        """

        :param address: Hostname or IP address of the SFTP server
        :param port: Port number
        :param user_name User name to be used to connect to the remote machine
        :param private_key: Private Key to be used to connect to the remote machine. The key has to be in PEM format and
          not OpenSSH format
        :param encryption_key Password for the private key itself, Defaults to empty string for no password
        :param target_resource_path TODO
        :param target_resource_type TODO
        """
        Credentials.__init__(self, SFTPCredentials.type_str)
        self.address = address
        self.port = port
        self.user_name = user_name
        self.private_key = private_key
        self.encryption_key = encryption_key
        self.target_resource_path = target_resource_path
        self.target_resource_type = target_resource_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(address: {self.address}, port: {self.port}, user_name: {self.user_name}, "
            f"private_key: ****, encryption_key: ****,"
            f"target_resource_path: '{self.target_resource_path}', target_resource_type: {self.target_resource_type!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["address"] = self.address
        base["port"] = self.port
        base["userName"] = self.user_name
        base["privateKey"] = self.private_key
        base["encryptionKey"] = self.encryption_key
        base["targetResourcePath"] = self.target_resource_path
        if self.target_resource_type is not None:
            base["targetResourceType"] = self.target_resource_type.value
        return base
